from __future__ import annotations

from typing import Any


class Progress:
    """
    Base class for progress reporting.  The default implementations do nothing.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    def update(self, advance: int = 1, **kwargs: float | int | str):
        """
        Update the progress bar.
        """
        pass

    def finish(self):
        """
        Finish and clean up this progress bar.  If the progresss bar is used as
        a context manager, this is automatically called on context exit.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.finish()



"""
Definitions and utilities related to LensKit data schemas.
"""

import warnings

import pandas as pd

from .types import AliasedColumn, Column

USER_COLUMN = "user_id"
ITEM_COLUMN = "item_id"
USER_COMPAT_COLUMN = AliasedColumn(USER_COLUMN, ["user"], warn=True)
ITEM_COMPAT_COLUMN = AliasedColumn(ITEM_COLUMN, ["item"], warn=True)


def column_name(col: Column) -> str:
    match col:
        case str(name):
            return name
        case AliasedColumn(name=name):
            return name
        case _:  # pragma: nocover
            raise TypeError(f"invalid column spec {col}")


def normalize_columns(df: pd.DataFrame, *columns: Column) -> pd.DataFrame:
    """
    Resolve column aliases to columns, pulling them out of the index if necessary.
    """

    for column in columns:
        name = column_name(column)
        if name in df.columns:
            continue
        elif name in df.index.names:
            df = df.reset_index(name)
            continue

        if not isinstance(column, AliasedColumn):
            raise KeyError("column %s not found", name)

        found = False
        for alias in column.compat_aliases:
            if alias in df.columns:
                found = True
                df = df.rename(columns={alias: name})
                if column.warn:
                    warnings.warn(f"found deprecated alias {alias} for {name}", DeprecationWarning)
                break
            elif alias in df.index.names:
                found = True
                df = df.reset_index(alias).rename(columns={alias: name})
                if column.warn:
                    warnings.warn(f"found deprecated alias {alias} for {name}", DeprecationWarning)
                break

        if not found:
            raise KeyError("column %s not found")

    return df



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Bias scoring model.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass

import numpy as np
import torch
from typing_extensions import Self, TypeAlias, overload

from lenskit.data import ID, Dataset, ItemList, QueryInput, RecQuery, UITuple, Vocabulary
from lenskit.pipeline.components import Component
from lenskit.stats import damped_mean

_logger = logging.getLogger(__name__)
Damping: TypeAlias = float | UITuple[float] | tuple[float, float]


@dataclass
class BiasModel:
    """
    User-item bias models learned from rating data.  The :class:`BiasScorer`
    class uses this model to score items in a pipeline; the model is reusable
    in other components that need user-item bias models.

    This implements the following model:

    .. math::
       b_{ui} = b_g + b_i + b_u

    where :math:`b_g` is the global bias (global mean rating), :math:`b_i` is
    item bias, and :math:`b_u` is the user bias.  With the provided damping
    values :math:`\\beta_{\\mathrm{u}}` and :math:`\\beta_{\\mathrm{i}}`, they
    are computed as follows:

    .. math::
       \\begin{align*}
       b_g & = \\frac{\\sum_{r_{ui} \\in R} r_{ui}}{|R|} &
       b_i & = \\frac{\\sum_{r_{ui} \\in R_i} (r_{ui} - b_g)}{|R_i| + \\beta_{\\mathrm{i}}} &
       b_u & = \\frac{\\sum_{r_{ui} \\in R_u} (r_{ui} - b_g - b_i)}{|R_u| + \\beta_{\\mathrm{u}}}
       \\end{align*}

    The damping values can be interpreted as the number of default (mean)
    ratings to assume *a priori* for each user or item, damping low-information
    users and items towards a mean instead of permitting them to take on extreme
    values based on few ratings.
    """

    damping: UITuple[float]
    "The mean damping terms."

    global_bias: float
    "The global bias term."

    items: Vocabulary | None = None
    "Vocabulary of items."
    item_biases: np.ndarray[int, np.dtype[np.float32]] | None = None
    "The item offsets (:math:`b_i` values)."

    users: Vocabulary | None = None
    "Vocabulary of users."
    user_biases: np.ndarray[int, np.dtype[np.float32]] | None = None
    "The user offsets (:math:`b_u` values)."

    @classmethod
    def learn(
        cls, data: Dataset, damping: Damping = 0.0, *, items: bool = True, users: bool = True
    ) -> Self:
        """
        Learn a bias model and its parameters from a dataset.

        Args:
            data:
                The dataset from which to learn the bias model.
            damping:
                Bayesian damping to apply to computed biases.  Either a number, to
                damp both user and item biases the same amount, or a (user,item)
                tuple providing separate damping values.
            items:
                Whether to compute item biases
            users:
                Whether to compute user biases
        """
        damping = UITuple.create(damping)

        _logger.info("building bias model for %d ratings", data.interaction_count)
        ratings = data.interaction_matrix("scipy", layout="coo", field="rating")
        nrows, ncols = ratings.shape  # type: ignore

        g_bias = float(np.mean(ratings.data))
        _logger.info("global mean: %.3f", g_bias)

        model = cls(damping, g_bias)

        centered = ratings.data - g_bias
        if np.allclose(centered, 0):
            _logger.warning("mean-centered ratings are all 0, bias probably meaningless")

        if items:
            counts = np.full(ncols, damping.item)
            sums = np.zeros(ncols)
            np.add.at(counts, ratings.col, 1)
            np.add.at(sums, ratings.col, centered)

            # store 0 offsets
            i_bias = np.zeros(ncols, dtype=np.float32)
            np.divide(sums, counts, out=i_bias, where=counts > 0)

            model.items = data.items.copy()
            model.item_biases = i_bias
            centered -= i_bias[ratings.col]
            _logger.info("computed biases for %d items", len(i_bias))

        if users:
            counts = np.full(nrows, damping.user)
            sums = np.zeros(nrows)
            np.add.at(counts, ratings.row, 1)
            np.add.at(sums, ratings.row, centered)

            u_bias = np.zeros(nrows, dtype=np.float32)
            np.divide(sums, counts, out=u_bias, where=counts > 0)

            model.users = data.users.copy()
            model.user_biases = u_bias
            _logger.info("computed biases for %d users", len(u_bias))

        return model

    @overload
    def compute_for_items(
        self,
        items: ItemList,
        user_id: ID | None = None,
        user_items: ItemList | None = None,
    ) -> tuple[np.ndarray[tuple[int], np.dtype[np.float32]], float]: ...
    @overload
    def compute_for_items(
        self,
        items: ItemList,
        *,
        bias: float,
    ) -> np.ndarray[tuple[int], np.dtype[np.float32]]: ...
    def compute_for_items(
        self,
        items: ItemList,
        user_id: ID | None = None,
        user_items: ItemList | None = None,
        *,
        bias: float | None = None,
    ):
        """
        Compute the personalized biases for a set of itemsm and optionally a
        user.  The user can be specified either by their identifier or by a list
        of ratings.

        Args:
            items:
                The items to score.
            user:
                The user identifier.
            user_items:
                The user's items, with ratings (takes precedence over ``user``
                if both are supplied).  If the supplied list does not have a
                ``rating`` field, it is ignored.
            bias:
                A pre-computed user bias.
        Returns:
            A tuple of the overall bias scores for the specified items and user,
            and the user's bias (needed to de-normalize scores efficiently
            later).  If a user bias is provided instead of user information,
            only the composite bias scores are returned.
        """
        n = len(items)
        scores = np.full(n, self.global_bias, dtype=np.float32)

        if self.item_biases is not None:
            assert self.items is not None
            idxes = items.numbers(vocabulary=self.items, missing="negative")
            mask = idxes >= 0
            scores[mask] += self.item_biases[idxes[mask]]

        if bias is not None:
            return scores + bias

        ratings = None
        if user_items is not None:
            ratings = user_items.field("rating")

        user_bias = 0.0
        if self.users is not None:
            assert self.user_biases is not None

            if ratings is not None:
                assert user_items is not None

                uoff = ratings - self.global_bias

                if self.item_biases is not None:
                    r_idxes = user_items.numbers(vocabulary=self.items, missing="negative")
                    r_mask = r_idxes >= 0
                    uoff[r_mask] -= self.item_biases[r_idxes[r_mask]]

                user_bias = damped_mean(uoff, self.damping.user)
                scores += user_bias

            elif user_id is not None:
                uno = self.users.number(user_id, missing="none")
                if uno is not None:
                    user_bias = self.user_biases[uno]
                    _logger.debug("using mean(user %s) = %.3f", user_id, user_bias)
                    scores += user_bias

        return scores, user_bias

    def transform_matrix(self, matrix: torch.Tensor):
        """
        Transform a sparse ratings matrix by subtracting biases.
        """
        if not matrix.is_sparse:
            raise TypeError("matrix is not sparse COO")

        indices = matrix.indices()
        unos = indices[0, :]
        inos = indices[1, :]
        values = matrix.values() - self.global_bias
        if self.item_biases is not None:
            values.subtract_(torch.from_numpy(self.item_biases)[inos])
        if self.user_biases is not None:
            values.subtract_(torch.from_numpy(self.user_biases)[unos])
        return torch.sparse_coo_tensor(indices, values, size=matrix.size())


class BiasScorer(Component):
    """
    A user-item bias rating prediction model.  This component uses
    :class:`BiasModel` to predict ratings for users and items.

    Args:
        items:
            Whether to compute item biases.
        users:
            Whether to compute user biases.
        damping:
            Bayesian damping to apply to computed biases.  Either a number, to
            damp both user and item biases the same amount, or a (user,item)
            tuple providing separate damping values.
    """

    IGNORED_CONFIG_FIELDS = ["user_damping", "item_damping"]

    users: bool
    items: bool
    damping: UITuple[float]
    "The configured offset damping levels."

    model_: BiasModel

    def __init__(
        self,
        items: bool = True,
        users: bool = True,
        damping: float | UITuple[float] | tuple[float, float] = 0.0,
        *,
        user_damping: float | None = None,
        item_damping: float | None = None,
    ):
        self.items = items
        self.users = users
        self.damping = UITuple.create(damping)

        if user_damping is not None or item_damping is not None:
            self.damping = UITuple(user=user_damping or 0.0, item=item_damping or 0.0)

        if self.damping.user < 0:
            raise ValueError("user damping must be non-negative")
        if self.damping.item < 0:
            raise ValueError("item damping must be non-negative")

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "model_")

    def train(self, data: Dataset):
        """
        Train the bias model on some rating data.

        Args:
            ratings:
                The training data (must have ratings).

        Returns:
            The trained bias object.
        """
        self.model_ = BiasModel.learn(data, self.damping, users=self.users, items=self.items)

    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Compute predictions for a user and items.  Unknown users and items are
        assumed to have zero bias.

        Args:
            query:
                The recommendation query.  If the query has an item list with
                ratings, those ratings are used to compute a new bias instead of
                using the user's recorded bias.
            items:
                The items to score.
        Returns:
            Scores for `items`.
        """
        query = RecQuery.create(query)

        scores, _bias = self.model_.compute_for_items(items, query.user_id, query.user_items)
        return ItemList(items, scores=scores)

    def __str__(self):
        return "Bias(ud={}, id={})".format(self.damping.user, self.damping.item)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Item-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import logging
import warnings

import numpy as np
import torch
from typing_extensions import Callable, Optional, TypeAlias, override

from lenskit import util
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.logging.progress import item_progress_handle, pbh_update
from lenskit.math.sparse import normalize_sparse_rows, safe_spmv
from lenskit.parallel import ensure_parallel_init
from lenskit.pipeline import Component, Trainable

_log = logging.getLogger(__name__)
MAX_BLOCKS = 1024


class ItemKNNScorer(Component, Trainable):
    """
    Item-item nearest-neighbor collaborative filtering feedback. This item-item
    implementation is based on the description of item-based CF by
    :cite:t:`deshpande:iknn` and hard-codes several design decisions found to
    work well in the previous Java-based LensKit code :cite:p:`lenskit-java`. In
    explicit-feedback mode, its output is equivalent to that of the Java
    version.

    Args:
        nnbrs:
            The maximum number of neighbors for scoring each item (``None`` for
            unlimited)
        min_nbrs:
            The minimum number of neighbors for scoring each item
        min_sim:
            Minimum similarity threshold for considering a neighbor.  Must be
            positive; if less than the smallest 32-bit normal (:math:`1.175
            \\times 10^{-38}`), is clamped to that value.
        save_nbrs:
            The number of neighbors to save per item in the trained model
            (``None`` for unlimited)
        feedback:
            The type of input data to use (explicit or implicit).  This affects
            data pre-processing and aggregation.
    """

    nnbrs: int
    min_nbrs: int = 1
    min_sim: float
    save_nbrs: int | None = None
    feedback: FeedbackType
    block_size: int

    items_: Vocabulary
    "Vocabulary of item IDs."
    item_means_: torch.Tensor | None
    "Mean rating for each known item."
    item_counts_: torch.Tensor
    "Number of saved neighbors for each item."
    sim_matrix_: torch.Tensor
    "Similarity matrix (sparse CSR tensor)."
    users_: Vocabulary
    "Vocabulary of user IDs."
    rating_matrix_: torch.Tensor
    "Normalized rating matrix to look up user ratings at prediction time."

    def __init__(
        self,
        nnbrs: int,
        min_nbrs: int = 1,
        min_sim: float = 1.0e-6,
        save_nbrs: int | None = None,
        feedback: FeedbackType = "explicit",
        block_size: int = 250,
    ):
        self.nnbrs = nnbrs
        self.min_nbrs = min_nbrs
        if self.min_nbrs is not None and self.min_nbrs < 1:
            self.min_nbrs = 1
        self.min_sim = min_sim
        self.save_nbrs = save_nbrs
        self.block_size = block_size

        self.feedback = feedback

        if self.min_sim < 0:
            _log.warning("item-item does not currently support negative similarities")
            warnings.warn("item-item does not currently support negative similarities")
        elif self.min_sim == 0:
            f4i = np.finfo("f4")
            _log.warning(
                "minimum similarity %e is too low, using %e", self.min_sim, f4i.smallest_normal
            )
            self.min_sim = float(f4i.smallest_normal)

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "items_")

    @override
    def train(self, data: Dataset):
        """
        Train a model.

        The model-training process depends on ``save_nbrs`` and ``min_sim``, but *not* on other
        algorithm parameters.

        Args:
            ratings:
                (user,item,rating) data for computing item similarities.
        """
        ensure_parallel_init()
        # Training proceeds in 2 steps:
        # 1. Normalize item vectors to be mean-centered and unit-normalized
        # 2. Compute similarities with pairwise dot products
        self._timer = util.Stopwatch()
        _log.info("training IKNN for %d users in %s feedback mode", data.item_count, self.feedback)

        _log.debug("[%s] beginning fit, memory use %s", self._timer, util.max_memory())

        field = "rating" if self.feedback == "explicit" else None
        init_rmat = data.interaction_matrix("torch", field=field)
        n_items = data.item_count
        _log.info(
            "[%s] made sparse matrix for %d items (%d ratings from %d users)",
            self._timer,
            n_items,
            len(init_rmat.values()),
            data.user_count,
        )
        _log.debug("[%s] made matrix, memory use %s", self._timer, util.max_memory())

        # we operate on *transposed* rating matrix: items on the rows
        rmat = init_rmat.transpose(0, 1).to_sparse_csr().to(torch.float64)

        if self.feedback == "explicit":
            rmat, means = normalize_sparse_rows(rmat, "center")
            if np.allclose(rmat.values(), 0.0):
                _log.warning("normalized ratings are zero, centering is not recommended")
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
        else:
            means = None
        _log.debug("[%s] centered, memory use %s", self._timer, util.max_memory())

        rmat, _norms = normalize_sparse_rows(rmat, "unit")
        _log.debug("[%s] normalized, memory use %s", self._timer, util.max_memory())

        _log.info("[%s] computing similarity matrix", self._timer)
        smat = self._compute_similarities(rmat)
        _log.debug("[%s] computed, memory use %s", self._timer, util.max_memory())

        _log.info(
            "[%s] got neighborhoods for %d of %d items",
            self._timer,
            np.sum(np.diff(smat.crow_indices()) > 0),
            n_items,
        )

        _log.info("[%s] computed %d neighbor pairs", self._timer, len(smat.col_indices()))

        self.items_ = data.items
        self.item_means_ = means
        self.item_counts_ = torch.diff(smat.crow_indices())
        self.sim_matrix_ = smat
        self.users_ = data.users
        self.rating_matrix_ = init_rmat
        _log.debug("[%s] done, memory use %s", self._timer, util.max_memory())

        return self

    def _compute_similarities(self, rmat: torch.Tensor):
        nitems, nusers = rmat.shape

        bs = max(self.block_size, nitems // MAX_BLOCKS)
        _log.debug("computing with effective block size %d", bs)
        with item_progress_handle("items", nitems) as pbh:
            smat = _sim_blocks(rmat.to(torch.float64), self.min_sim, self.save_nbrs, bs, pbh)

        return smat.to(torch.float32)

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)
        _log.debug("predicting %d items for user %s", len(items), query.user_id)

        ratings = query.user_items
        if ratings is None:
            if query.user_id is None:
                raise ValueError("cannot recommend without without either user ID or items")

            upos = self.users_.number(query.user_id, missing=None)
            if upos is None:
                _log.debug("user %s missing, returning empty predictions", query.user_id)
                return ItemList(items, scores=np.nan)
            row = self.rating_matrix_[upos]  # type: ignore
            ratings = ItemList(
                item_nums=row.indices()[0], rating=row.values(), vocabulary=self.items_
            )

        # set up rating array
        # get rated item positions & limit to in-model items
        ri_pos = ratings.numbers(format="torch", vocabulary=self.items_, missing="negative")
        ri_mask = ri_pos >= 0
        ri_vpos = ri_pos[ri_mask]
        n_valid = len(ri_vpos)
        _log.debug("user %s: %d of %d rated items in model", query.user_id, n_valid, len(ratings))

        if self.feedback == "explicit":
            ri_vals = ratings.field("rating", "torch")
            if ri_vals is None:
                raise RuntimeError("explicit-feedback scorer must have ratings")
            ri_vals = ri_vals[ri_mask].to(torch.float64)
        else:
            ri_vals = torch.full((n_valid,), 1.0, dtype=torch.float64)

        # mean-center the rating array
        if self.feedback == "explicit":
            assert self.item_means_ is not None
            ri_vals -= self.item_means_[ri_vpos]

        # now compute the predictions
        if self.feedback == "explicit":
            sims = _predict_weighted_average(
                self.sim_matrix_, (self.min_nbrs, self.nnbrs), ri_vals, ri_vpos
            )
            sims += self.item_means_
        else:
            sims = _predict_sum(self.sim_matrix_, (self.min_nbrs, self.nnbrs), ri_vals, ri_vpos)

        # and prepare the output
        scores = torch.full((len(items),), np.nan, dtype=sims.dtype)
        out_nums = items.numbers("torch", vocabulary=self.items_, missing="negative")
        out_good = out_nums >= 0
        scores[out_good] = sims[out_nums[out_nums >= 0]]
        results = ItemList(items, scores=scores)

        _log.debug(
            "user %s: predicted for %d of %d items",
            query.user_id,
            int(torch.isfinite(scores).sum()),
            len(items),
        )

        return results

    def __str__(self):
        return "ItemItem(nnbrs={}, msize={})".format(self.nnbrs, self.save_nbrs)


@torch.jit.ignore  # type: ignore
def _msg(level, msg):
    # type: (int, str) -> None
    _log.log(level, msg)


@torch.jit.script
def _sim_row(
    item: int, matrix: torch.Tensor, row: torch.Tensor, min_sim: float, max_nbrs: Optional[int]
) -> tuple[int, torch.Tensor, torch.Tensor]:
    nitems, nusers = matrix.shape
    if len(row.indices()) == 0:
        return 0, torch.zeros((0,), dtype=torch.int32), torch.zeros((0,), dtype=torch.float64)

    # _item_dbg(item, f"comparing item with {row.indices().shape[1]} users")
    # _item_dbg(item, f"row norm {torch.linalg.vector_norm(row.values()).item()}")
    row = row.to_dense()
    sim = safe_spmv(matrix, row.to(torch.float64))
    sim[item] = 0

    mask = sim >= min_sim
    cols = torch.nonzero(mask)[:, 0].to(torch.int32)
    vals = sim[mask]
    # _item_dbg(item, f"found {len(vals)} acceptable similarities")
    assert len(cols) == torch.sum(mask)

    if max_nbrs is not None and max_nbrs > 0 and max_nbrs < vals.shape[0]:
        # _item_dbg(item, "truncating similarities")
        vals, cis = torch.topk(vals, max_nbrs, sorted=False)
        cols = cols[cis]
        order = torch.argsort(cols)
        cols = cols[order]
        vals = vals[order]

    return len(cols), cols, torch.clamp(vals, -1, 1)


@torch.jit.script
def _sim_block(
    matrix: torch.Tensor, start: int, end: int, min_sim: float, max_nbrs: Optional[int], pbh: str
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    "Compute a single block of the similarity matrix"
    bsize = end - start

    counts = torch.zeros(bsize, dtype=torch.int32)
    columns = []
    values = []

    for i in range(start, end):
        c, cs, vs = _sim_row(i, matrix, matrix[i], min_sim, max_nbrs)
        counts[i - start] = c
        columns.append(cs)
        values.append(vs)
        pbh_update(pbh, 1)

    return counts, torch.cat(columns), torch.cat(values).to(torch.float32)


@torch.jit.script
def _sim_blocks(
    matrix: torch.Tensor, min_sim: float, max_nbrs: Optional[int], block_size: int, pbh: str
) -> torch.Tensor:
    "Compute the similarity matrix with blocked matrix-matrix multiplies"
    nitems, nusers = matrix.shape

    jobs: list[torch.jit.Future[tuple[torch.Tensor, torch.Tensor, torch.Tensor]]] = []

    for start in range(0, nitems, block_size):
        end = min(start + block_size, nitems)
        jobs.append(torch.jit.fork(_sim_block, matrix, start, end, min_sim, max_nbrs, pbh))  # type: ignore

    counts = [torch.tensor([0], dtype=torch.int32)]
    columns = []
    values = []

    for job in jobs:
        cts, cis, vs = job.wait()
        counts.append(cts)
        columns.append(cis)
        values.append(vs)

    c_cat = torch.cat(counts)
    crow_indices = torch.cumsum(c_cat, 0, dtype=torch.int32)
    assert len(crow_indices) == nitems + 1
    col_indices = torch.cat(columns)
    c_values = torch.cat(values)
    assert crow_indices[nitems] == len(col_indices)
    assert crow_indices[nitems] == len(c_values)

    return torch.sparse_csr_tensor(
        crow_indices=crow_indices,
        col_indices=col_indices,
        values=c_values,
        size=(nitems, nitems),
    )


def _predict_weighted_average(
    model: torch.Tensor,
    nrange: tuple[int, int],
    rate_v: torch.Tensor,
    rated: torch.Tensor,
) -> torch.Tensor:
    "Weighted average prediction function"
    nitems, _ni = model.shape
    assert nitems == _ni
    min_nbrs, max_nbrs = nrange

    # we proceed rating-by-rating, and accumulate results
    scores = torch.zeros(nitems)
    t_sims = torch.zeros(nitems)
    counts = torch.zeros(nitems, dtype=torch.int32)
    # these store the similarities and values for neighbors, so we can un-count
    nbr_sims = torch.empty((nitems, max_nbrs))
    nbr_vals = torch.empty((nitems, max_nbrs))
    # and this stores the smallest similarity so far for each item
    nbr_min = torch.full((nitems,), torch.finfo().max)

    for i, iidx in enumerate(rated):
        row = model[int(iidx)]
        row_is = row.indices()[0]
        row_vs = row.values()
        assert row_is.shape == row_vs.shape

        row_avs = torch.abs(row_vs)
        fast = counts[row_is] < max_nbrs

        # save the fast-path items
        if torch.any(fast):
            ris_fast = row_is[fast]
            vs_fast = row_vs[fast]
            avs_fast = row_avs[fast]
            vals_fast = vs_fast * rate_v[i]
            nbr_sims[ris_fast, counts[ris_fast]] = vs_fast
            nbr_vals[ris_fast, counts[ris_fast]] = vals_fast
            counts[ris_fast] += 1
            t_sims[ris_fast] += avs_fast
            scores[ris_fast] += vals_fast
            nbr_min[ris_fast] = torch.minimum(nbr_min[ris_fast], vs_fast)

        # skip early if we're done
        if torch.all(fast):
            continue

        # now we have the slow-path items
        slow = torch.logical_not(fast)
        ris_slow = row_is[slow]
        rvs_slow = row_vs[slow]
        # which slow items might actually need an update?
        exc = rvs_slow > nbr_min[ris_slow]
        if not torch.any(exc):
            continue

        ris_slow = ris_slow[exc]
        rvs_slow = rvs_slow[exc]

        # this is brute-force linear search for simplicity right now
        # for each, find the neighbor that's the smallest:
        min_sims, mins = torch.min(nbr_sims[ris_slow], dim=1)
        assert torch.all(min_sims < rvs_slow)

        # now we need to update values: add in new and remove old
        min_vals = nbr_vals[ris_slow, mins]
        ravs_slow = row_avs[slow][exc]
        slow_vals = rvs_slow * rate_v[i]
        t_sims[ris_slow] += ravs_slow - min_sims.abs()
        scores[ris_slow] += slow_vals - min_vals
        # and save
        nbr_sims[ris_slow, mins] = ravs_slow
        nbr_vals[ris_slow, mins] = slow_vals
        # and now we need to update the saved minimums
        nm_sims, _nm_is = torch.min(nbr_sims[ris_slow], dim=1)
        nbr_min[ris_slow] = nm_sims

    # compute averages for items that pass match the threshold
    mask = counts >= min_nbrs
    scores[mask] /= t_sims[mask]
    scores[torch.logical_not(mask)] = torch.nan

    return scores


def _predict_sum(
    model: torch.Tensor,
    nrange: tuple[int, int],
    rate_v: torch.Tensor,
    rated: torch.Tensor,
) -> torch.Tensor:
    "Sum-of-similarities prediction function"
    nitems, _ni = model.shape
    assert nitems == _ni
    min_nbrs, max_nbrs = nrange
    _msg(logging.DEBUG, f"sum-scoring with {len(rated)} items")

    # we proceed rating-by-rating, and accumulate results
    t_sims = torch.zeros(nitems)
    counts = torch.zeros(nitems, dtype=torch.int32)
    nbr_sims = torch.zeros((nitems, max_nbrs))
    # and this stores the smallest similarity so far for each item
    nbr_min = torch.full((nitems,), torch.finfo().max)

    for i, iidx in enumerate(rated):
        iidx = int(iidx)
        row = model[iidx]
        row_is = row.indices()[0]
        row_vs = row.values()
        assert row_is.shape == row_vs.shape

        fast = counts[row_is] < max_nbrs

        # save the fast-path items
        if torch.any(fast):
            ris_fast = row_is[fast]
            vs_fast = row_vs[fast]
            nbr_sims[ris_fast, counts[ris_fast]] = vs_fast
            counts[ris_fast] += 1
            t_sims[ris_fast] += vs_fast
            nbr_min[ris_fast] = torch.minimum(nbr_min[ris_fast], vs_fast)

        # skip early if we're done
        if torch.all(fast):
            continue

        # now we have the slow-path items
        slow = torch.logical_not(fast)
        ris_slow = row_is[slow]
        rvs_slow = row_vs[slow]
        # which slow items might actually need an update?
        exc = rvs_slow > nbr_min[ris_slow]
        if not torch.any(exc):
            continue

        ris_slow = ris_slow[exc]
        rvs_slow = rvs_slow[exc]

        # this is brute-force linear search for simplicity right now
        # for each, find the neighbor that's the smallest:
        min_sims, mins = torch.min(nbr_sims[ris_slow], dim=1)

        # now we need to update values: add in new and remove old
        # anywhere our new neighbor is grater than smallest, replace smallest
        t_sims[ris_slow] -= min_sims
        t_sims[ris_slow] += rvs_slow
        # and save
        nbr_sims[ris_slow, mins] = rvs_slow
        # save the minimums
        nm_sims, _nm_is = torch.min(nbr_sims[ris_slow], dim=1)
        nbr_min[ris_slow] = nm_sims

    # compute averages for items that pass match the threshold
    t_sims[counts < min_nbrs] = torch.nan

    return t_sims


AggFun: TypeAlias = Callable[
    [
        torch.Tensor,
        tuple[int, int],
        torch.Tensor,
        torch.Tensor,
    ],
    torch.Tensor,
]



"""
Functions for working with bulk per-user data.

.. note::

    This is a provisional API to enable incremental forward development while
    we develop more flexible abstractions for collections of data indexed by
    user or other keys.
"""

from collections.abc import Mapping
from typing import Iterator, cast, overload

import pandas as pd

from .items import ItemList
from .schemas import (
    ITEM_COMPAT_COLUMN,
    USER_COLUMN,
    USER_COMPAT_COLUMN,
    column_name,
    normalize_columns,
)
from .types import ID, Column


def dict_to_df(data: Mapping[ID, ItemList | None], *, column: str = USER_COLUMN) -> pd.DataFrame:
    """
    Convert a dictionary mapping user IDs to item lists into a data frame.
    Missing item lists are excluded.

    Args:
        data:
            The dictionary of data.
        column:
            The column, to support dictionaries mapped by things other than user
            IDs.
    """

    df = pd.concat(
        {u: il.to_df(numbers=False) for (u, il) in data.items() if il is not None},
        names=[column],
    )
    df = df.reset_index(column)
    df = df.reset_index(drop=True)
    return df


def dict_from_df(
    df: pd.DataFrame, *, column: Column = USER_COMPAT_COLUMN, item_col: Column = ITEM_COMPAT_COLUMN
) -> dict[ID, ItemList]:
    """
    Convert a dictionary mapping user IDs to item lists into a data frame.

    Args:
        df:
            The data frame.
        column:
            The column, to support dictionaries mapped by things other than user IDs.
    """
    df = normalize_columns(df, column, item_col)
    return {u: ItemList.from_df(udf) for (u, udf) in df.groupby(column_name(column))}  # type: ignore


def group_df(df: pd.DataFrame, *, column: Column = USER_COMPAT_COLUMN, item_col=ITEM_COMPAT_COLUMN):
    """
    Group a data frame by a specified column, possibly checking for and
    normalizing the names of other columns as well.  The default options group
    by ``user_id`` and require an ``item_id`` column, allowing the compatibility
    aliases ``user`` and ``item``.
    """
    df = normalize_columns(df, column, item_col)
    col = column_name(column)
    return df.groupby(col)


@overload
def count_item_lists(data: Mapping[ID, ItemList | None]) -> int: ...
@overload
def count_item_lists(
    data: pd.DataFrame | Mapping[ID, ItemList | None],
    *,
    column: Column = USER_COMPAT_COLUMN,
) -> int: ...
def count_item_lists(
    data: pd.DataFrame | Mapping[ID, ItemList | None],
    *,
    column: Column = USER_COMPAT_COLUMN,
) -> int:
    if isinstance(data, pd.DataFrame):
        data = normalize_columns(data, column)
        return data[column_name(column)].nunique()
    else:
        return len(data)


@overload
def iter_item_lists(
    data: Mapping[ID, ItemList | None],
) -> Iterator[tuple[ID, ItemList | None]]: ...
@overload
def iter_item_lists(
    data: pd.DataFrame | Mapping[ID, ItemList | None],
    *,
    column: Column = USER_COMPAT_COLUMN,
    item_col=ITEM_COMPAT_COLUMN,
) -> Iterator[tuple[ID, ItemList | None]]: ...
def iter_item_lists(
    data: pd.DataFrame | Mapping[ID, ItemList | None],
    *,
    column: Column = USER_COMPAT_COLUMN,
    item_col=ITEM_COMPAT_COLUMN,
) -> Iterator[tuple[ID, ItemList | None]]:
    """
    Iterate over item lists identified by keys.  When the input is a data frame,
    the column names may be specified; the default options group by ``user_id``
    and require an ``item_id`` column, allowing the compatibility aliases
    ``user`` and ``item``.
    """
    if isinstance(data, pd.DataFrame):
        for key, df in group_df(data, column=column, item_col=item_col):
            yield cast(ID, key), ItemList.from_df(df)

    else:
        yield from data.items()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"Definition of the component interfaces."

# pyright: strict
from __future__ import annotations

import inspect
import json
from abc import abstractmethod
from importlib import import_module
from types import FunctionType
from typing import Callable, ClassVar, Generic, ParamSpec, TypeAlias

from typing_extensions import Any, Protocol, Self, TypeVar, override, runtime_checkable

from lenskit.data.dataset import Dataset

from .types import Lazy

P = ParamSpec("P")
T = TypeVar("T")
# COut is only return, so Component[U] can be assigned to Component[T] if U ≼ T.
COut = TypeVar("COut", covariant=True)
PipelineFunction: TypeAlias = Callable[..., COut]


@runtime_checkable
class Configurable(Protocol):  # pragma: nocover
    """
    Interface for configurable objects such as pipeline components with settings
    or hyperparameters.  A configurable object supports two operations:

    * saving its configuration with :meth:`get_config`.
    * creating a new instance from a saved configuration with the class method
      :meth:`from_config`.

    An object must implement both of these methods to be considered
    configurable.  Components extending the :class:`Component` automatically
    have working versions of these methods if they define their constructor
    parameters and fields appropriately.

    .. note::

        Configuration data should be JSON-compatible (strings, numbers, etc.).
    """

    @classmethod
    def from_config(cls, cfg: dict[str, Any]) -> Self:
        """
        Reinstantiate this component from configuration values.
        """
        raise NotImplementedError()

    def get_config(self) -> dict[str, object]:
        """
        Get this component's configured hyperparameters.
        """
        raise NotImplementedError()


@runtime_checkable
class Trainable(Protocol):  # pragma: nocover
    """
    Interface for components that can learn parameters from training data. It
    supports trainingand checking if a component has already been trained.
    Trained components need to be picklable.

    .. note::

        Trainable components must also implement ``__call__``.

    .. note::

        A future LensKit version will add support for extracting model
        parameters a la Pytorch's ``state_dict``, but this capability was not
        ready for 2025.1.
    """

    @property
    def is_trained(self) -> bool:
        """
        Check if this model has already been trained.
        """
        raise NotImplementedError()

    def train(self, data: Dataset) -> None:
        """
        Train the pipeline component to learn its parameters from a training
        dataset.

        Args:
            data:
                The training dataset.
            retrain:
                If ``True``, retrain the model even if it has already been
                trained.
        """
        raise NotImplementedError()


@runtime_checkable
class ParameterContainer(Protocol):  # pragma: nocover
    def get_params(self) -> dict[str, object]:
        """
        Protocol for components with parameters that can be extracted, saved,
        and re-loaded.

        LensKit components that learn parameters from training data should both
        implement this method and work when pickled and unpickled.  Pickling is
        sometimes used for convenience, but parameter / state dictionaries allow
        serializing wtih tools like ``safetensors``.

        .. todo::

            This protocol is not yet used for anything.

        Args:
            include_caches:
                Whether the parameter dictionary should include ephemeral
                caching structures only used for runtime performance
                optimizations.

        Returns:
            The model's parameters, as a dictionary from names to parameter data
            (usually arrays, tensors, etc.).
        """
        raise NotImplementedError()

    def load_params(self, params: dict[str, object]) -> None:
        """
        Reload model state from parameters saved via :meth:`get_params`.
        """
        raise NotImplementedError()


class Component(Configurable, Generic[COut]):
    """
    Base class for pipeline component objects.  Any component that is not just a
    function should extend this class.

    Components are :class:`Configurable`.  The base class provides default
    implementations of :meth:`get_config` and :meth:`from_config` that inspect
    the constructor arguments and instance variables to automatically provide
    configuration support.  By default, all constructor parameters will be
    considered configuration parameters, and their values will be read from
    instance variables of the same name. Components can also define
    :data:`EXTRA_CONFIG_FIELDS` and :data:`IGNORED_CONFIG_FIELDS` class
    variables to modify this behavior. Missing attributes are silently ignored.

    To work as components, derived classes also need to implement a ``__call__``
    method to perform their operations.
    """

    EXTRA_CONFIG_FIELDS: ClassVar[list[str]] = []
    """
    Names of instance variables that should be included in the configuration
    dictionary even though they do not correspond to named constructor
    arguments.

    .. note::

        This is rarely needed, and usually needs to be coupled with ``**kwargs``
        in the constructor to make the resulting objects constructible.
    """

    IGNORED_CONFIG_FIELDS: ClassVar[list[str]] = []
    """
    Names of constructor parameters that should be excluded from the
    configuration dictionary.
    """

    @override
    def get_config(self) -> dict[str, object]:
        """
        Get the configuration by inspecting the constructor and instance
        variables.
        """
        sig = inspect.signature(self.__class__)
        names = list(sig.parameters.keys()) + self.EXTRA_CONFIG_FIELDS
        params: dict[str, Any] = {}
        for name in names:
            if name not in self.IGNORED_CONFIG_FIELDS and hasattr(self, name):
                params[name] = getattr(self, name)

        return params

    @override
    @classmethod
    def from_config(cls, cfg: dict[str, Any]) -> Self:
        """
        Create a class from the specified construction.  Configuration elements
        are passed to the constructor as keywrod arguments.
        """
        return cls(**cfg)

    @abstractmethod
    def __call__(self, **kwargs: Any) -> COut:
        """
        Run the pipeline's operation and produce a result.  This is the key
        method for components to implement.

        .. note::

            Due to limitations of Python's type model, derived classes will have
            a type error overriding this method when using strict type checking,
            because it is very cumbersome (if not impossible) to propagate
            parameter names and types through to a base class.  The solution is
            to either use basic type checking for implementations, or to disable
            the typechecker on the ``__call__`` signature definition.
        """
        ...

    def __repr__(self) -> str:
        params = json.dumps(self.get_config(), indent=2)
        return f"<{self.__class__.__name__} {params}>"


def instantiate_component(
    comp: str | type | FunctionType, config: dict[str, Any] | None
) -> Callable[..., object]:
    """
    Utility function to instantiate a component given its class, function, or
    string representation.
    """
    if isinstance(comp, str):
        mname, oname = comp.split(":", 1)
        mod = import_module(mname)
        comp = getattr(mod, oname)

    # make the type checker happy
    assert not isinstance(comp, str)

    if isinstance(comp, FunctionType):
        return comp
    elif issubclass(comp, Configurable):
        if config is None:
            config = {}
        return comp.from_config(config)  # type: ignore
    else:
        return comp()  # type: ignore


def fallback_on_none(input: T | None, fallback: Lazy[T]) -> T:
    if input is not None:
        return input
    else:
        return fallback.get()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Inspect the Python environment setup.  We provide this in LensKit so that people
can inspect (and report) their setups.
"""

import os
import platform
import re
import sys
from importlib.metadata import distributions, version

import threadpoolctl


def inspect_platform():
    print("Python version:", platform.python_version())
    print("Platform:", platform.platform())
    print("Python location:", sys.executable)


def inspect_version():
    print("LensKit version:", version("lenskit"))


def inspect_compute():
    import numpy as np
    import torch

    print("\nNumPy version:", np.__version__)
    print("PyTorch version:", torch.__version__)
    print("PyTorch backends:")
    print("    cpu:", torch.backends.cpu.get_cpu_capability())
    for mod in [torch.cuda, torch.backends.mkl, torch.backends.mps]:
        if mod.is_available():
            stat = "available"
        elif hasattr(mod, "is_built") and mod.is_built():
            stat = "unavailable"
        else:
            stat = "absent"
        name = mod.__name__.split(".")[-1]
        print(f"    {name}: {stat}")

    print("\nThreading layers:")
    for i, pool in enumerate(threadpoolctl.threadpool_info(), 1):
        print(f"    Backend {i}:")
        for k, v in pool.items():
            print("      {}: {}".format(k, v))


def inspect_env():
    print("\nEnvironment variables:")
    for k, v in os.environ.items():
        if re.match(r"^(LK_|OMP_|NUMBA_|MKL_|TORCH_)", k):
            print(f"    {k}: {v}")
    print("\nPython search paths:")
    for path in sys.path:
        print(f"    {path}")
    print("\nExecutable search paths:")
    paths = os.environ["PATH"].split(os.pathsep)
    for path in paths:
        print(f"    {path}")


def inspect_packages():
    dists = sorted(distributions(), key=lambda d: d.name)
    print("\nInstalled packages ({}):".format(len(dists)))
    for dist in dists:
        print("    {:32s}  {:>10s}".format(dist.name, dist.version))


def main():
    inspect_platform()
    inspect_version()
    inspect_compute()
    inspect_env()
    inspect_packages()


if __name__ == "__main__":
    main()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic data types used in data representations.
"""

# pyright: strict
from __future__ import annotations

import sys
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NamedTuple,
    Sequence,
    TypeAlias,
    TypeVar,
    cast,
)

import numpy as np
import pandas as pd

FeedbackType: TypeAlias = Literal["explicit", "implicit"]
"Types of feedback supported."

CoreID: TypeAlias = int | str | bytes
"Core (non-NumPy) identifier types."
NPID: TypeAlias = np.integer[Any] | np.str_ | np.bytes_ | np.object_
"NumPy entity identifier types."
ID: TypeAlias = CoreID | NPID
"Allowable identifier types."
IDArray: TypeAlias = np.ndarray[tuple[int], np.dtype[NPID]]
"NumPy arrays of identifiers."
IDSequence: TypeAlias = Sequence[ID] | IDArray | "pd.Series[CoreID]"
"Sequences of identifiers."

T = TypeVar("T")


@dataclass
class AliasedColumn:
    name: str
    "The column name."
    compat_aliases: list[str] = field(default_factory=list)
    "A list of aliases for the column."
    warn: bool = False
    "Whether to warn when using an alias."


Column: TypeAlias = str | AliasedColumn


if TYPE_CHECKING or sys.version_info >= (3, 11):

    class UITuple(NamedTuple, Generic[T]):
        """
        Tuple of (user, item) data, typically for configuration and similar
        purposes.
        """

        user: T
        "User data."
        item: T
        "Item data."

        @classmethod
        def create(cls, x: UITuple[T] | tuple[T, T] | T) -> UITuple[T]:
            """
            Create a user-item tuple from a tuple or data.  If a single value
            is provided, it is used for both user and item.
            """
            if isinstance(x, UITuple):
                return cast(UITuple[T], x)
            elif isinstance(x, (tuple, list)):
                u, i = cast(tuple[T, T], x)
                return UITuple(u, i)
            elif isinstance(x, dict):
                return UITuple(x["user"], x["item"])
            else:
                return UITuple(x, x)
else:

    class UITuple(NamedTuple):
        """
        Tuple of (user, item) data, typically for configuration and similar
        purposes.
        """

        user: Any
        "User data."
        item: Any
        "Item data."

        @classmethod
        def create(cls, x: UITuple | tuple[Any, Any] | Any) -> UITuple:
            """
            Create a user-item tuple from a tuple or data.  If a single value
            is provided, it is used for both user and item.
            """
            if isinstance(x, UITuple):
                return x
            elif isinstance(x, tuple):
                u, i = x
                return UITuple(u, i)
            else:
                return UITuple(x, x)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging

import numpy as np
import torch
from typing_extensions import override

from lenskit.basic import BiasModel
from lenskit.data import Dataset, ItemList
from lenskit.data.types import UITuple
from lenskit.logging.progress import item_progress_handle, pbh_update
from lenskit.math.solve import solve_cholesky
from lenskit.parallel.chunking import WorkChunks
from lenskit.types import RNGInput

from ._common import ALSBase, TrainContext, TrainingData

_log = logging.getLogger(__name__)


class BiasedMFScorer(ALSBase):
    """
    Biased matrix factorization trained with alternating least squares
    :cite:p:`zhouLargeScaleParallelCollaborative2008`.  This is a
    prediction-oriented algorithm suitable for explicit feedback data, using the
    alternating least squares approach to compute :math:`P` and :math:`Q` to
    minimize the regularized squared reconstruction error of the ratings matrix.

    See the base class :class:`ALSBase` for documentation on the estimated
    parameters you can extract from a trained model.

    Args:
        features:
            The number of features to train.
        epochs:
            The number of iterations to train.
        reg:
            The regularization factor; can also be a tuple ``(ureg, ireg)`` to
            specify separate user and item regularization terms.
        damping:
            Damping term for the bias model.
        rng:
            Random number seed or generator.
    """

    timer = None

    damping: UITuple[float]

    bias_: BiasModel

    def __init__(
        self,
        features: int,
        *,
        epochs: int = 10,
        reg: float | tuple[float, float] = 0.1,
        damping: float | UITuple[float] | tuple[float, float] = 5.0,
        rng: RNGInput = None,
        save_user_features: bool = True,
    ):
        super().__init__(
            features,
            epochs=epochs,
            reg=reg,
            rng=rng,
            save_user_features=save_user_features,
        )
        self.damping = UITuple.create(damping)

    @property
    def logger(self):
        return _log

    @override
    def prepare_data(self, data: Dataset):
        # transform ratings using offsets
        rmat = data.interaction_matrix("torch", layout="coo", field="rating")

        _log.info("[%s] normalizing ratings", self.timer)
        self.bias_ = BiasModel.learn(data, self.damping)
        rmat = self.bias_.transform_matrix(rmat)

        rmat = rmat.to_sparse_csr()
        assert not torch.any(torch.isnan(rmat.values()))
        return TrainingData.create(data.users, data.items, rmat)

    @override
    def initial_params(self, nrows: int, ncols: int, rng: np.random.Generator) -> torch.Tensor:
        mat = rng.standard_normal((nrows, ncols))
        mat /= np.linalg.norm(mat, axis=1).reshape((nrows, 1))
        mat = torch.from_numpy(mat)
        return mat

    @override
    def als_half_epoch(self, epoch: int, context: TrainContext):
        chunks = WorkChunks.create(context.nrows)
        with item_progress_handle(f"epoch {epoch} {context.label}s", total=context.nrows) as pbh:
            return _train_update_fanout(context, chunks, pbh)

    @override
    def new_user_embedding(
        self, user_num: int | None, items: ItemList
    ) -> tuple[torch.Tensor, float | None]:
        inums = items.numbers("torch", vocabulary=self.items_, missing="negative")
        mask = inums >= 0
        ratings = items.field("rating", "torch")
        assert ratings is not None

        biases, u_bias = self.bias_.compute_for_items(items, None, items)
        biases = torch.from_numpy(biases)
        ratings = ratings - biases

        ri_val = ratings[mask].to(torch.float64)

        u_feat = _train_bias_row_cholesky(inums[mask], ri_val, self.item_features_, self.reg.user)
        return u_feat, u_bias

    @override
    def finalize_scores(
        self, user_num: int | None, items: ItemList, user_bias: float | None
    ) -> ItemList:
        scores = items.scores()
        assert scores is not None

        if user_bias is None:
            if user_num is not None and self.bias_.user_biases is not None:
                user_bias = self.bias_.user_biases[user_num]
            else:
                user_bias = 0.0
        assert user_bias is not None

        biases = self.bias_.compute_for_items(items, bias=user_bias)
        scores = scores + biases

        return ItemList(items, scores=scores)

    def __str__(self):
        return "als.BiasedMFScorer(features={}, regularization={})".format(self.features, self.reg)


@torch.jit.script
def _train_solve_row(
    cols: torch.Tensor,
    vals: torch.Tensor,
    this: torch.Tensor,
    other: torch.Tensor,
    regI: torch.Tensor,
) -> torch.Tensor:
    (nui,) = cols.shape
    nf = other.shape[1]

    if nui == 0:
        return torch.zeros(nf)

    M = other[cols, :]
    MMT = M.T @ M
    # assert MMT.shape[0] == ctx.n_features
    # assert MMT.shape[1] == ctx.n_features
    A = MMT + regI * nui
    V = M.T @ vals
    # and solve
    return solve_cholesky(A, V)


@torch.jit.script
def _train_update_rows(ctx: TrainContext, start: int, end: int, pbh: str) -> torch.Tensor:
    result = ctx.left[start:end, :].clone()

    for i in range(start, end):
        row = ctx.matrix[i]
        (n,) = row.shape
        if n == 0:
            continue

        cols = row.indices()[0]
        vals = row.values().type(ctx.left.type())

        V = _train_solve_row(cols, vals, ctx.left, ctx.right, ctx.regI)
        result[i - start] = V
        pbh_update(pbh, 1)

    return result


@torch.jit.script
def _train_update_fanout(ctx: TrainContext, chunking: WorkChunks, pbh: str) -> float:
    if ctx.nrows <= 50:
        # at 50 rows, we run sequentially
        M = _train_update_rows(ctx, 0, ctx.nrows, pbh)
        sqerr = torch.norm(ctx.left - M)
        ctx.left[:, :] = M
        return sqerr.item()

    results: list[tuple[int, int, torch.jit.Future[torch.Tensor]]] = []
    for start in range(0, ctx.nrows, chunking.chunk_size):
        end = min(start + chunking.chunk_size, ctx.nrows)
        results.append((start, end, torch.jit.fork(_train_update_rows, ctx, start, end, pbh)))  # type: ignore

    sqerr = torch.tensor(0.0)
    for start, end, r in results:
        M = r.wait()
        diff = (ctx.left[start:end, :] - M).ravel()
        sqerr += torch.dot(diff, diff)
        ctx.left[start:end, :] = M

    return sqerr.sqrt().item()


def _train_bias_row_cholesky(
    items: torch.Tensor, ratings: torch.Tensor, other: torch.Tensor, reg: float
) -> torch.Tensor:
    """
    Args:
        items: the item IDs the user has rated
        ratings: the user's (normalized) ratings for those items
        other: the item-feature matrix
        reg: the regularization term
    Returns:
        the user-feature vector (equivalent to V in the current Cholesky code)
    """
    (nui,) = items.shape
    nf = other.shape[1]

    if nui == 0:
        return torch.zeros(nf)

    M = other[items, :]
    regI = torch.eye(nf, device=other.device) * reg
    MMT = M.T @ M
    A = MMT + regI * len(items)

    V = M.T @ ratings
    x = solve_cholesky(A, V)

    return x



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Vocabularies of IDs, tags, etc.
"""

# pyright: basic
from __future__ import annotations

from hashlib import sha1
from typing import Hashable, Iterable, Iterator, Literal, Sequence, overload
from warnings import warn

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray

from lenskit.diagnostics import DataWarning

from .types import ID, IDArray, IDSequence


class Vocabulary:
    """
    Vocabularies of terms, tags, entity IDs, etc. for the LensKit data model.

    This class supports bidirectional mappings between key-like data and
    congiguous nonnegative integer indices.  Its key use is to facilitate the
    user and item ID vocabularies in :class:`~lenskit.data.Dataset`, but it can
    also be used for things like item tags.

    It is currently a wrapper around :class:`pandas.Index`, but supports the
    ability to add additional vocabulary terms after the vocabulary has been
    created.  New terms do not change the index positions of previously-known
    identifiers.
    """

    name: str | None
    "The name of the vocabulary (e.g. “user”, “item”)."
    _index: pd.Index
    _hashes: dict[int, str]

    def __init__(
        self,
        keys: IDSequence | pd.Index | Iterable[ID] | None = None,
        name: str | None = None,
    ):
        self.name = name
        if keys is None:
            keys = pd.Index()
        elif isinstance(keys, pd.Index):
            if not pd.unique:
                raise ValueError("vocabulary index must be unique")
        elif isinstance(keys, np.ndarray) or isinstance(keys, list) or isinstance(keys, pd.Series):
            keys = pd.Index(np.unique(keys))  # type: ignore
        else:
            keys = pd.Index(sorted(set(keys)))  # type: ignore

        self._index = keys.rename(name) if name is not None else keys
        self._hashes = {}

    @property
    def index(self) -> pd.Index:
        "The property as a Pandas index."
        return self._index

    @property
    def size(self) -> int:
        "Current vocabulary size."
        return len(self._index)

    @overload
    def number(self, term: object, missing: Literal["error"] = "error") -> int: ...
    @overload
    def number(self, term: object, missing: Literal["none"] | None) -> int | None: ...
    def number(
        self, term: object, missing: Literal["error", "none"] | None = "error"
    ) -> int | None:
        "Look up the number for a vocabulary term."
        try:
            num = self._index.get_loc(term)
            assert isinstance(num, int)
            return num
        except KeyError as e:
            if missing == "error":
                raise e
            else:
                return None

    def numbers(
        self, terms: Sequence[Hashable] | ArrayLike, missing: Literal["error", "negative"] = "error"
    ) -> np.ndarray[int, np.dtype[np.int32]]:
        "Look up the numbers for an array of terms or IDs."
        nums = np.require(self._index.get_indexer_for(terms), dtype=np.int32)
        if missing == "error" and np.any(nums < 0):
            raise KeyError()
        return nums

    def term(self, num: int) -> object:
        """
        Look up the term with a particular number.  Negative indexing is **not** supported.
        """
        if num < 0:
            raise IndexError("negative numbers not supported")
        return self._index[num]

    def terms(self, nums: list[int] | NDArray[np.integer] | pd.Series | None = None) -> IDArray:
        """
        Get a list of terms, optionally for an array of term numbers.

        Args:
            nums:
                The numbers (indices) for of terms to retrieve.  If ``None``,
                returns all terms.

        Returns:
            The terms corresponding to the specified numbers, or the full array
            of terms (in order) if ``nums=None``.
        """
        if nums is not None:
            nums = np.asarray(nums, dtype=np.int32)
            if np.any(nums < 0):
                raise IndexError("negative numbers not supported")
            return self._index[nums].values
        else:
            return self._index.values

    def id(self, num: int) -> object:
        "Alias for :meth:`term`  for greater readability for entity ID vocabularies."
        return self.term(num)

    def ids(self, nums: list[int] | NDArray[np.integer] | pd.Series | None = None) -> IDArray:
        "Alias for :meth:`terms` for greater readability for entity ID vocabularies."
        return self.terms(nums)

    def add_terms(self, terms: list[Hashable] | ArrayLike):
        arr = np.unique(terms)  # type: ignore
        nums = self.numbers(arr, missing="negative")
        fresh = arr[nums < 0]
        self._index = pd.Index(np.concatenate([self._index.values, fresh]), name=self.name)

    def copy(self) -> Vocabulary:
        """
        Return a (cheap) copy of this vocabulary.  It retains the same mapping,
        but will not be updated if the original vocabulary has new terms added.
        However, since new terms are always added to the end, it will be
        compatible with the original vocabulary for all terms recorded at the
        time of the copy.

        This method is useful for saving known vocabularies in model training.
        """
        return Vocabulary(self._index)

    def compatible_with_numbers_from(self, other: Vocabulary | None) -> bool:
        """
        Check if this vocabulary is compatible with numbers from another
        vocabulary.  They are compatible if the other vocabulary is no longer
        than this vocabulary, and the common prefix has identical IDs.

        Args:
            other:
                The other vocabulary.

        Returns:
            ``True`` the same IDs will produce the same numbers from both
            vocabularies.
        """
        if other is None:
            return False
        if self is other:
            return True

        if len(self) < len(other):
            return False

        h1 = self._hash(len(other))
        h2 = other._hash()
        return h1 == h2

    def _hash(self, length: int | None = None) -> str:
        if length is None or length > len(self._index):
            length = len(self._index)

        h = self._hashes.get(length, None)
        if h is None:
            hasher = sha1(usedforsecurity=False)
            arr = self._index.values[:length]
            if arr.dtype == np.object_:
                # we have to hash each object
                warn(f"slowly hashing IDs (dtype {arr.dtype})", DataWarning, 3)
                for i in arr:
                    hasher.update(repr(i).encode())
            else:
                hasher.update(memoryview(arr))

            h = hasher.hexdigest()
            self._hashes[length] = h

        return h

    def __eq__(self, other: Vocabulary) -> bool:  # noqa: F821
        return self.size == other.size and bool(np.all(self.index == other.index))

    def __contains__(self, key: object) -> bool:
        return key in self._index

    def __iter__(self) -> Iterator[object]:
        return iter(self._index.values)

    def __len__(self) -> int:
        return self.size

    def __array__(self, dtype=None) -> np.ndarray:
        return self._index.values.__array__(dtype)

    def __str__(self) -> str:
        if self.name:
            return f"Vocabulary of {self.size} {self.name} terms"
        else:
            return f"Vocabulary of {self.size} terms"

    def __repr__(self) -> str:
        if self.name:
            return f"<Vocabulary(name={self.name}, size={self.size})>"
        else:
            return f"<Vocabulary(size={self.size})>"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from collections.abc import Mapping
from typing import Any, Iterator

from .config import PipelineMeta


class PipelineState(Mapping[str, Any]):
    """
    Full results of running a pipeline.  A pipeline state is a dictionary
    mapping node names to their results; it is implemented as a separate class
    instead of directly using a dictionary to allow data to be looked up by node
    aliases in addition to original node names (and to be read-only).

    Client code will generally not construct this class directly.

    Args:
        state:
            The pipeline state to wrap.  The state object stores a reference to
            this dictionary.
        aliases:
            Dictionary of node aliases.
        default:
            The name of the default node (whose data should be returned by
            :attr:`default` ).
        meta:
            The metadata for the pipeline generating this state.
    """

    _state: dict[str, Any]
    _aliases: dict[str, str]
    _default: str | None = None
    meta: PipelineMeta | None
    """
    Pipeline metadata.
    """

    def __init__(
        self,
        state: dict[str, Any] | None = None,
        aliases: dict[str, str] | None = None,
        default: str | None = None,
        meta: PipelineMeta | None = None,
    ) -> None:
        self._state = state if state is not None else {}
        self._aliases = aliases if aliases is not None else {}
        self._default = default
        self.meta = meta
        if default is not None and default not in self:
            raise ValueError("default node is not in state or aliases")

    @property
    def default(self) -> Any:
        """
        Return the data from of the default node (typically the last node run).

        Returns:
            The data associated with the default node.

        Raises:
            ValueError: if there is no specified default node.
        """
        if self._default is not None:
            return self[self._default]
        else:
            raise ValueError("pipeline state has no default value")

    @property
    def default_node(self) -> str | None:
        "Return the name of the default node (typically the last node run)."
        return self._default

    def __len__(self):
        return len(self._state)

    def __contains__(self, key: object) -> bool:
        if key in self._state:
            return True
        if key in self._aliases:
            return self._aliases[key] in self
        else:
            return False

    def __getitem__(self, key: str) -> Any:
        if key in self._state:
            return self._state[key]
        elif key in self._aliases:
            return self[self._aliases[key]]
        else:
            raise KeyError(f"pipeline node <{key}>")

    def __iter__(self) -> Iterator[str]:
        return iter(self._state)

    def __str__(self) -> str:
        return f"<PipelineState with {len(self)} nodes>"

    def __repr__(self) -> str:
        return f"<PipelineState with nodes {set(self._state.keys())}>"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pydantic models for pipeline configuration and serialization support.
"""

# pyright: strict
from __future__ import annotations

import base64
import pickle
from collections import OrderedDict
from hashlib import sha256
from types import FunctionType
from typing import Literal

from pydantic import BaseModel, Field, JsonValue, ValidationError
from typing_extensions import Any, Optional, Self

from .components import Configurable
from .nodes import ComponentNode, InputNode
from .types import type_string


class PipelineConfig(BaseModel):
    """
    Root type for serialized pipeline configuration.  A pipeline config contains
    the full configuration, components, and wiring for the pipeline, but does
    not contain the
    """

    meta: PipelineMeta
    "Pipeline metadata."
    inputs: list[PipelineInput] = Field(default_factory=list)
    "Pipeline inputs."
    defaults: dict[str, str] = Field(default_factory=dict)
    "Default pipeline wirings."
    components: OrderedDict[str, PipelineComponent] = Field(default_factory=OrderedDict)
    "Pipeline components, with their configurations and wiring."
    aliases: dict[str, str] = Field(default_factory=dict)
    "Pipeline node aliases."
    literals: dict[str, PipelineLiteral] = Field(default_factory=dict)
    "Literals"


class PipelineMeta(BaseModel):
    """
    Pipeline metadata.
    """

    name: str | None = None
    "The pipeline name."
    version: str | None = None
    "The pipeline version."
    hash: str | None = None
    """
    The pipeline configuration hash.  This is optional, particularly when
    hand-crafting pipeline configuration files.
    """


class PipelineInput(BaseModel):
    name: str
    "The name for this input."
    types: Optional[set[str]]
    "The list of types for this input."

    @classmethod
    def from_node(cls, node: InputNode[Any]) -> Self:
        if node.types is not None:
            types = {type_string(t) for t in node.types}
        else:
            types = None

        return cls(name=node.name, types=types)


class PipelineComponent(BaseModel):
    code: str
    """
    The path to the component's implementation, either a class or a function.
    This is a Python qualified path of the form ``module:name``.
    """

    config: dict[str, object] | None = Field(default=None)
    """
    The component configuration.  If not provided, the component will be created
    with its default constructor parameters.
    """

    inputs: dict[str, str] | list[str] = Field(default_factory=dict)
    """
    The component's input wirings, mapping input names to node names.  For
    certain meta-nodes, it is specified as a list instead of a dict.
    """

    @classmethod
    def from_node(cls, node: ComponentNode[Any], mapping: dict[str, str] | None = None) -> Self:
        if mapping is None:
            mapping = {}

        comp = node.component
        if isinstance(comp, FunctionType):
            ctype = comp
        else:
            ctype = comp.__class__

        code = f"{ctype.__module__}:{ctype.__qualname__}"

        config = comp.get_config() if isinstance(comp, Configurable) else None

        return cls(
            code=code,
            config=config,
            inputs={n: mapping.get(t, t) for (n, t) in node.connections.items()},
        )


class PipelineLiteral(BaseModel):
    """
    Literal nodes represented in the pipeline.
    """

    encoding: Literal["json", "base85"]
    value: JsonValue

    @classmethod
    def represent(cls, data: Any) -> Self:
        try:
            return cls(encoding="json", value=data)
        except ValidationError:
            # data is not basic JSON values, so let's pickle it
            dbytes = pickle.dumps(data)
            return cls(encoding="base85", value=base64.b85encode(dbytes).decode("ascii"))

    def decode(self) -> Any:
        "Decode the represented literal."
        match self.encoding:
            case "json":
                return self.value
            case "base85":
                assert isinstance(self.value, str)
                return pickle.loads(base64.b85decode(self.value))


def hash_config(config: BaseModel) -> str:
    """
    Compute the hash of a configuration model.
    """
    json = config.model_dump_json(exclude_none=True)
    h = sha256()
    h.update(json.encode())
    return h.hexdigest()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Data abstractions and data set access.
"""

from __future__ import annotations

from .collection import GenericKey, ItemListCollection, UserIDKey
from .convert import from_interactions_df
from .dataset import Dataset, FieldError
from .items import ItemList
from .movielens import load_movielens, load_movielens_df
from .mtarray import MTArray, MTFloatArray, MTGenericArray, MTIntArray
from .query import QueryInput, RecQuery
from .types import ID, NPID, FeedbackType, UITuple
from .vocab import Vocabulary

__all__ = [
    "Dataset",
    "FieldError",
    "from_interactions_df",
    "ID",
    "NPID",
    "UITuple",
    "FeedbackType",
    "ItemList",
    "ItemListCollection",
    "UserIDKey",
    "GenericKey",
    "load_movielens",
    "load_movielens_df",
    "MTArray",
    "MTFloatArray",
    "MTGenericArray",
    "MTIntArray",
    "Vocabulary",
    "RecQuery",
    "QueryInput",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"Data check functions for LensKit."

# pyright: strict
from __future__ import annotations

from typing import Any, Literal, Protocol, TypeVar, overload

import numpy as np
import pyarrow as pa
from numpy.typing import NDArray


class HasShape(Protocol):
    @property
    def shape(self) -> tuple[int, ...]: ...


A = TypeVar("A", bound=HasShape)
NPT = TypeVar("NPT", bound=np.generic)


@overload
def check_1d(
    arr: A,
    size: int | None = None,
    *,
    label: str = "array",
    error: Literal["raise"] = "raise",
) -> A: ...
@overload
def check_1d(
    arr: HasShape,
    size: int | None = None,
    *,
    error: Literal["return"],
) -> bool: ...
def check_1d(
    arr: A,
    size: int | None = None,
    *,
    label: str = "array",
    error: Literal["raise", "return"] = "raise",
) -> bool | A:
    """
    Check that an array is one-dimensional, optionally checking that it has the
    expected length.

    This check function has 2 modes:

    *   If ``error="raise"`` (the default), it will raise a :class:`TypeError`
        if the array shape is incorrect, and return the array otherwise.
    *   If ``error="return"``, it will return ``True`` or ``False`` depending on
        whether the size is correct.

    Args:
        arr:
            The array to check.
        size:
            The expected size of the array. If unspecified, this function simply
            checks that the array is 1-dimensional, but does not check the size
            of that dimension.
        label:
            A label to use in the exception message.
        error:
            The behavior when an array fails the test.

    Returns:
        The array, if ``error="raise"`` and the array passes the check, or a
        boolean indicating whether it passes the check.

    Raises:
        TypeError: if ``error="raise"`` and the array fails the check.
    """
    if size is None and len(arr.shape) > 1:
        if error == "raise":
            raise TypeError(f"{label} must be 1D (has shape {arr.shape})")
        else:
            return False
    elif size is not None and arr.shape != (size,):
        if error == "raise":
            raise TypeError(f"{label} has incorrect shape (found {arr.shape}, expected {size})")
        else:
            return False

    if error == "raise":
        return arr
    else:
        return True


@overload
def check_type(
    arr: NDArray[Any],
    *types: type[NPT],
    label: str = "array",
    error: Literal["raise"] = "raise",
) -> NDArray[NPT]: ...
@overload
def check_type(
    arr: NDArray[Any],
    *types: type[NPT],
    error: Literal["return"],
) -> bool: ...
def check_type(
    arr: NDArray[Any],
    *types: type[NPT],
    label: str = "array",
    error: Literal["raise", "return"] = "raise",
) -> bool | NDArray[Any]:
    """
    Check that an array array is of an acceptable type.

    This check function has 2 modes:

    *   If ``error="raise"`` (the default), it will raise a :class:`TypeError`
        if the array shape is incorrect, and return the array otherwise.
    *   If ``error="return"``, it will return ``True`` or ``False`` depending on
        whether the size is correct.

    Args:
        arr:
            The array to check.
        types:
            The acceptable types for the array.
        label:
            A label to use in the exception message.
        error:
            The behavior when an array fails the test.

    Returns:
        The array, if ``error="raise"`` and the array passes the check, or a
        boolean indicating whether it passes the check.

    Raises:
        TypeError: if ``error="raise"`` and the array fails the check.
    """
    if issubclass(arr.dtype.type, types):
        if error == "raise":
            return arr
        else:
            return True
    elif error == "raise":
        raise TypeError(f"{label} has incorrect type {arr.dtype} (allowed: {types})")
    else:
        return False


def array_is_null(array: object) -> bool:
    if array is None:
        return True
    if isinstance(array, pa.Array) and not np.any(array.is_valid()):
        return True

    return False



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Test utilities for LKPY tests.
"""

import os
import os.path
from contextlib import contextmanager
from typing import Literal

import numpy as np
import pandas as pd
import scipy.sparse as sps
from pyprojroot import here

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
import pytest
from hypothesis import assume

from lenskit.basic import PopScorer, SoftmaxRanker
from lenskit.batch import recommend
from lenskit.data import Dataset, ItemList, ItemListCollection, UserIDKey, from_interactions_df
from lenskit.data.lazy import LazyDataset
from lenskit.data.movielens import load_movielens, load_movielens_df
from lenskit.math.sparse import torch_sparse_from_scipy
from lenskit.pipeline import RecPipelineBuilder
from lenskit.splitting import TTSplit, simple_test_pair

ml_test_dir = here("data/ml-latest-small")
ml_100k_zip = here("data/ml-100k.zip")

ml_test: Dataset = LazyDataset(lambda: load_movielens(ml_test_dir))


@pytest.fixture(scope="session")
def ml_ratings():
    """
    Fixture to load the test MovieLens ratings as a data frame. To use this,
    just include it as a parameter in your test::

        def test_thing_with_data(ml_ratings: pd.DataFrame):
            ...

    .. note::
        This is imported in ``conftest.py`` so it is always available in LensKit tests.
    """
    yield load_movielens_df(ml_test_dir)


@pytest.fixture(scope="session")
def ml_ds(ml_ratings: pd.DataFrame):
    """
    Fixture to load the MovieLens test dataset.  To use this, just include it as
    a parameter in your test::

        def test_thing_with_data(ml_ds: Dataset):
            ...

    .. note::
        This is imported in ``conftest.py`` so it is always available in LensKit tests.
    """
    yield from_interactions_df(ml_ratings)


@pytest.fixture
def ml_100k():
    """
    Fixture to load the MovieLens 100K dataset (currently as a data frame).  It skips
    the test if the ML100K data is not available.
    """
    if not ml_100k_zip.exists():
        pytest.skip("ML100K data not available")
    yield load_movielens_df(ml_100k_zip)


@pytest.fixture(scope="session")
def demo_recs(ml_ds: Dataset) -> tuple[TTSplit, ItemListCollection[UserIDKey]]:
    """
    A demo set of train, test, and recommendation data.
    """
    rng = np.random.default_rng(42)
    split = simple_test_pair(ml_ds, f_rates=0.5, rng=rng)

    builder = RecPipelineBuilder()
    builder.scorer(PopScorer())
    builder.ranker(SoftmaxRanker(500))
    pipe = builder.build()
    pipe.train(split.train)

    recs = recommend(pipe, list(split.test.keys()), 500, n_jobs=1, rng=rng)
    return split, recs


@contextmanager
def set_env_var(var, val):
    "Set an environment variable & restore it."
    old_val = os.environ.get(var, None)
    try:
        if val is None:
            if old_val is not None:
                del os.environ[var]
        else:
            os.environ[var] = val
        yield
    finally:
        if old_val is not None:
            os.environ[var] = old_val
        elif val is not None:
            del os.environ[var]


@st.composite
def coo_arrays(
    draw,
    shape=None,
    dtype=nph.floating_dtypes(endianness="=", sizes=[32, 64]),
    elements=st.floats(-1e6, 1e6, allow_nan=False, allow_infinity=False, width=32),
) -> sps.coo_array:
    if shape is None:
        shape = st.tuples(st.integers(1, 100), st.integers(1, 100))

    if isinstance(shape, st.SearchStrategy):
        shape = draw(shape)

    if not isinstance(shape, tuple):
        shape = shape, shape
    rows, cols = shape
    if isinstance(rows, st.SearchStrategy):
        rows = draw(rows)
    if isinstance(cols, st.SearchStrategy):
        cols = draw(cols)

    mask = draw(nph.arrays(np.bool_, (rows, cols)))
    # at least one nonzero value
    assume(np.any(mask))
    nnz = int(np.sum(mask))

    ris, cis = np.nonzero(mask)

    vs = draw(
        nph.arrays(dtype, nnz, elements=elements),
    )

    return sps.coo_array((vs, (ris, cis)), shape=(rows, cols))


@st.composite
def sparse_arrays(draw, *, layout="csr", **kwargs):
    if isinstance(layout, list):
        layout = st.sampled_from(layout)
    if isinstance(layout, st.SearchStrategy):
        layout = draw(layout)

    M: sps.coo_array = draw(coo_arrays(**kwargs))

    match layout:
        case "csr":
            return M.tocsr()
        case "csc":
            return M.tocsc()
        case "coo":
            return M
        case _:
            raise ValueError(f"invalid layout {layout}")


@st.composite
def sparse_tensors(draw, *, layout="csr", **kwargs):
    if isinstance(layout, list):
        layout = st.sampled_from(layout)
    if isinstance(layout, st.SearchStrategy):
        layout = draw(layout)

    M: sps.coo_array = draw(coo_arrays(**kwargs))
    return torch_sparse_from_scipy(M, layout)  # type: ignore


@st.composite
def scored_lists(
    draw: st.DrawFn,
    *,
    n: int | tuple[int, int] | st.SearchStrategy[int] = st.integers(0, 1000),
    scores: st.SearchStrategy[float] | Literal["gaussian"] | None = None,
) -> ItemList:
    """
    Hypothesis generator that produces scored lists.
    """
    if isinstance(n, st.SearchStrategy):
        n = draw(n)
    elif isinstance(n, tuple):
        n = draw(st.integers(*n))

    ids = np.arange(1, n + 1, dtype=np.int32)
    if scores == "gaussian":
        seed = draw(st.integers(0))
        rng = np.random.default_rng(seed)
        xs = np.exp(rng.normal(size=n))
    else:
        xs = draw(
            nph.arrays(nph.floating_dtypes(endianness="=", sizes=[32, 64]), n, elements=scores)
        )
    return ItemList(item_ids=ids, scores=xs)


jit_enabled = True
if "NUMBA_DISABLE_JIT" in os.environ:
    jit_enabled = False
if os.environ.get("PYTORCH_JIT", None) == "0":
    jit_enabled = False

wantjit = pytest.mark.skipif(not jit_enabled, reason="JIT required")



import numpy as np
from typing_extensions import override

from lenskit.data import ItemList

from ._base import ListMetric, RankingMetricBase


class RecipRank(ListMetric, RankingMetricBase):
    """
    Compute the reciprocal rank :cite:p:`trec5-confusion` of the first relevant
    item in a list of recommendations.  Taking the mean of this metric over the
    recommendation lists in a run yields the MRR (mean reciprocal rank).

    Let :math:`\\kappa` denote the 1-based rank of the first relevant item in
    :math:`L`, with :math:`\\kappa=\\infty` if none of the first :math:`k` items
    in :math:`L` are relevant; then the reciprocal rank is :math:`1 / \\kappa`.
    If no elements are relevant, the reciprocal rank is therefore 0.
    :cite:t:`deshpande:iknn` call this the “reciprocal hit rate”.
    """

    @property
    def label(self):
        if self.k is not None:
            return f"RecipRank@{self.k}"
        else:
            return "RecipRank"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        if len(test) == 0:
            return np.nan

        recs = self.truncate(recs)
        items = recs.ids()
        good = np.isin(items, test.ids())
        (npz,) = np.nonzero(good)
        if len(npz):
            return 1.0 / (npz[0] + 1.0)
        else:
            return 0.0



import datetime as dt
from typing import Sequence, overload

import numpy as np

from lenskit.data import Dataset, from_interactions_df
from lenskit.data.collection import ItemListCollection

from .split import TTSplit


@overload
def split_global_time(data: Dataset, time: str | dt.datetime) -> TTSplit: ...
@overload
def split_global_time(data: Dataset, time: Sequence[str | dt.datetime]) -> list[TTSplit]: ...
def split_global_time(
    data: Dataset, time: str | dt.datetime | Sequence[str | dt.datetime]
) -> TTSplit | list[TTSplit]:
    """
    Global temporal train-test split.  This splits a data set into train/test
    pairs using a single global timestamp.  When given multiple timestamps, it
    will return multiple splits, where split :math:`i` has training data from
    before :math:`t_i` and testing data on or after :math:`t_i` and before
    :math:`t_{i+1}` (the last split has no upper bound on the testing data).

    Args:
        data:
            The dataset to split.
        time:
            Time or sequence of times at which to split.  Strings must be in ISO
            format.

    Returns:
        The data splits.
    """
    if isinstance(time, (str, int, float, dt.datetime)):
        times = [_make_time(time)]
        rv = "single"
    else:
        times = [_make_time(t) for t in time]
        rv = "sequence"

    matrix = data.interaction_log("pandas", fields="all", original_ids=True)
    if "timestamp" not in matrix:
        raise RuntimeError("temporal split requires timestamp")

    ts_col = matrix["timestamp"]
    ts_col = np.asarray(ts_col)

    if ts_col.dtype.kind in ("i", "u", "f"):
        times = [t.timestamp() for t in times]

    results = []
    for i, t in enumerate(times):
        mask = ts_col >= t

        train = matrix[~mask]
        if i + 1 < len(times):
            test = matrix[mask & (ts_col < times[i + 1])]
        else:
            test = matrix[mask]

        train_ds = from_interactions_df(train)
        test_ilc = ItemListCollection.from_df(test, ["user_id"])
        results.append(TTSplit(train_ds, test_ilc))

    if rv == "sequence":
        return results
    else:
        assert len(results) == 1
        return results[0]


def _make_time(t: int | float | str | dt.datetime) -> dt.datetime:
    if isinstance(t, (int, float)):
        return dt.datetime.fromtimestamp(t)
    elif isinstance(t, str):
        return dt.datetime.fromisoformat(t)
    else:
        return t



import numpy as np
from numpy.typing import NDArray
from typing_extensions import Callable, TypeAlias, override

from lenskit.data import ItemList

from ._base import ListMetric, RankingMetricBase

Discount: TypeAlias = Callable[[NDArray[np.number]], NDArray[np.float64]]


class NDCG(ListMetric, RankingMetricBase):
    """
    Compute the normalized discounted cumulative gain :cite:p:`ndcg`.

    Discounted cumultative gain is computed as:

    .. math::
        \\begin{align*}
        \\mathrm{DCG}(L,u) & = \\sum_{i=1}^{|L|} \\frac{r_{ui}}{d(i)}
        \\end{align*}

    Unrated items are assumed to have a utility of 0; if no rating values are
    provided in the truth frame, item ratings are assumed to be 1.

    This is then normalized as follows:

    .. math::
        \\begin{align*}
        \\mathrm{nDCG}(L, u) & = \\frac{\\mathrm{DCG}(L,u)}{\\mathrm{DCG}(L_{\\mathrm{ideal}}, u)}
        \\end{align*}

    Args:
        k:
            The maximum recommendation list length to consider (longer lists are
            truncated).
        discount:
            The discount function to use.  The default, base-2 logarithm, is the
            original function used by :cite:t:`ndcg`.
        gain:
            The field on the test data to use for gain values.  If ``None`` (the
            default), all items present in the test data have a gain of 1.  If set
            to a string, it is the name of a field (e.g. ``'rating'``).  In all
            cases, items not present in the truth data have a gain of 0.
    """

    discount: Discount
    gain: str | None

    def __init__(
        self, k: int | None = None, *, discount: Discount = np.log2, gain: str | None = None
    ):
        super().__init__(k=k)
        self.discount = discount
        self.gain = gain

    @property
    def label(self):
        if self.k is not None:
            return f"NDCG@{self.k}"
        else:
            return "NDCG"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        items = recs.ids()

        if self.gain:
            gains = test.field(self.gain, "pandas", index="ids")
            if gains is None:
                raise KeyError(f"test items have no field {self.gain}")
            scores = gains.reindex(items, fill_value=0).values
            if self.k:
                gains = gains.nlargest(n=self.k)
            else:
                gains = gains.sort_values(ascending=False)
            ideal = array_dcg(np.require(gains.values, np.float32), self.discount)
        else:
            scores = np.zeros_like(items, dtype=np.float32)
            scores[np.isin(items, test.ids())] = 1.0
            n = len(test)
            if self.k and self.k < n:
                n = self.k
            ideal = fixed_dcg(n, self.discount)

        realized = array_dcg(np.require(scores, np.float32), self.discount)
        return realized / ideal


def array_dcg(scores: NDArray[np.number], discount: Discount = np.log2):
    """
    Compute the Discounted Cumulative Gain of a series of recommended items with rating scores.
    These should be relevance scores; they can be :math:`{0,1}` for binary relevance data.

    This is not a true top-N metric, but is a utility function for other metrics.

    Args:
        scores:
            The utility scores of a list of recommendations, in recommendation order.
        discount:
            the rank discount function.  Each item's score will be divided the discount of its rank,
            if the discount is greater than 1.

    Returns:
        double: the DCG of the scored items.
    """
    scores = np.nan_to_num(scores)
    ranks = np.arange(1, len(scores) + 1)
    disc = discount(ranks)
    np.maximum(disc, 1, out=disc)
    np.reciprocal(disc, out=disc)
    return np.dot(scores, disc)


def fixed_dcg(n: int, discount: Discount = np.log2):
    """
    Compute the Discounted Cumulative Gain of a fixed number of items with
    relevance 1.
    """

    ranks = np.arange(1, n + 1)
    disc = discount(ranks)
    disc = np.maximum(disc, 1)
    disc = np.reciprocal(disc)
    return np.sum(disc)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Iterable, Literal, Mapping, TypeAlias

from lenskit.data import ID, GenericKey, ItemList, ItemListCollection, UserIDKey
from lenskit.logging import item_progress
from lenskit.parallel import invoker
from lenskit.pipeline import Pipeline
from lenskit.util import Stopwatch

from ._results import BatchResults

_log = logging.getLogger(__name__)

ItemSource: TypeAlias = None | Literal["test-items"]
"""
Types of items that can be returned.
"""


@dataclass
class InvocationSpec:
    """
    Specification for a single pipeline invocation, to record one or more
    pipeline component outputs for a test user.
    """

    name: str
    "A name for this invocation."
    components: dict[str, str]
    "The names of pipeline components to measure and return, mapped to their output names."
    items: ItemSource = None
    "The target or candidate items (if any) to provide to the recommender."
    extra_inputs: dict[str, Any] = field(default_factory=dict)
    "Additional inputs to pass to the pipeline."


class BatchPipelineRunner:
    """
    Apply a pipeline to a collection of test users.

    Argss:
        pipeline:
            The pipeline to evaluate.
        n_jobs:
            The number of parallel processes to use, or ``None`` for the
            default (defined by :func:`lenskit.parallel.config.initialize`).
    """

    n_jobs: int | None
    invocations: list[InvocationSpec]

    def __init__(self, *, n_jobs: int | None = None):
        self.n_jobs = n_jobs
        self.invocations = []

    def add_invocation(self, inv: InvocationSpec):
        self.invocations.append(inv)

    def predict(self, component: str = "rating-predictor", *, output: str = "predictions"):
        """
        Request the batch run to generate test item scores or rating predictins.

        Args:
            component:
                The name of the rating predictor component to run.
            output:
                The name of the results in the output dictionary.
        """
        self.add_invocation(InvocationSpec("predict-ratings", {component: output}, "test-items"))

    def recommend(
        self, component: str = "recommender", *, output: str = "recommendations", **extra: Any
    ):
        """
        Request the batch run to generate recomendations.

        Args:
            component:
                The name of the recommender component to run.
            output:
                The name of the results in the output dictionary.
            extra:
                Extra inputs to the recommender. A common option is ``n``, the
                number of recommendations to return (a default may be baked into
                the pipeline).
        """
        self.add_invocation(InvocationSpec("recommend", {component: output}, extra_inputs=extra))

    def run(
        self,
        pipeline: Pipeline,
        test_data: ItemListCollection[GenericKey]
        | Mapping[ID, ItemList]
        | Iterable[ID | GenericKey],
    ) -> BatchResults:
        """
        Run the pipeline and return its results.

        Args:
            test_data:
                The collection of test data, as an ItemListCollection, a mapping
                of user IDs to test data, or as a sequence of item IDs for
                recommendation.

        Returns:
            The results, as a nested dictionary.  The outer dictionary maps
            component output names to inner dictionaries of result data.  These
            inner dictionaries map user IDs to
        """
        if isinstance(test_data, ItemListCollection):
            test_iter = test_data
            key_type = test_data.key_type
            n_users = len(test_data)
        elif isinstance(test_data, Mapping):
            key_type = UserIDKey
            test_iter = ((UserIDKey(k), v) for (k, v) in test_data.items())
            n_users = len(test_data)
        else:
            key_type = UserIDKey
            test_data = list(test_data)
            test_iter = ((_ensure_key(k), None) for k in test_data)
            n_users = len(test_data)

        n_users = len(test_data)
        _log.info("running pipeline %s for %d queries", pipeline.name, n_users)
        _log.info("pipeline configuration hash: %s", pipeline.config_hash())

        with (
            invoker((pipeline, self.invocations), _run_pipeline, n_jobs=self.n_jobs) as worker,
            item_progress("Recommending", n_users) as progress,
        ):
            # release our reference, will sometimes free the pipeline memory in this process
            del pipeline
            results = BatchResults(key_type)
            timer = Stopwatch()
            for key, outs in worker.map(test_iter):
                for cn, cr in outs.items():
                    results.add_result(cn, key, cr)
                progress.update()
            timer.stop()

            rate = timer.elapsed() / n_users
            _log.info(
                "finished running for %d users in %s (%.1fms/user)",
                n_users,
                timer.elapsed() * 1000,
                rate,
            )

        return results


def _ensure_key(key: ID | tuple[ID, ...]):
    if isinstance(key, tuple):
        return key
    else:
        return UserIDKey(key)


def _run_pipeline(
    ctx: tuple[Pipeline, list[InvocationSpec]],
    req: tuple[GenericKey, ItemList] | ID,
) -> tuple[GenericKey, dict[str, object]]:
    pipeline, invocations = ctx
    key, test_items = req

    result = {}

    _log.debug("running pipeline %s for request %s", pipeline.name, key)
    for inv in invocations:
        inputs: dict[str, Any] = {}
        if hasattr(key, "user_id"):
            inputs["query"] = key.user_id  # type: ignore
        match inv.items:
            case "test-items":
                inputs["items"] = test_items

        inputs.update(inv.extra_inputs)

        nodes = inv.components.keys()
        outs = pipeline.run_all(*nodes, **inputs)
        for cname, oname in inv.components.items():
            result[oname] = outs[cname]

    return key, result  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from abc import ABC, abstractmethod

import numpy as np
import torch
from typing_extensions import Iterator, NamedTuple, Self, override

from lenskit import util
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.types import UITuple
from lenskit.logging import item_progress
from lenskit.parallel.config import ensure_parallel_init
from lenskit.pipeline import Component, Trainable
from lenskit.types import RNGInput
from lenskit.util.random import random_generator


class TrainContext(NamedTuple):
    """
    Context object for one half of an ALS training operation.
    """

    label: str
    matrix: torch.Tensor
    left: torch.Tensor
    right: torch.Tensor
    reg: float
    nrows: int
    ncols: int
    embed_size: int
    regI: torch.Tensor

    @classmethod
    def create(
        cls, label: str, matrix: torch.Tensor, left: torch.Tensor, right: torch.Tensor, reg: float
    ) -> TrainContext:
        nrows, ncols = matrix.shape
        lnr, embed_size = left.shape
        assert lnr == nrows
        assert right.shape == (ncols, embed_size)
        regI = torch.eye(embed_size, dtype=left.dtype, device=left.device) * reg
        return TrainContext(label, matrix, left, right, reg, nrows, ncols, embed_size, regI)


class TrainingData(NamedTuple):
    """
    Data for training the ALS model.
    """

    users: Vocabulary
    "User ID mapping."
    items: Vocabulary
    "Item ID mapping."
    ui_rates: torch.Tensor
    "User-item rating matrix."
    iu_rates: torch.Tensor
    "Item-user rating matrix."

    @property
    def n_users(self):
        return len(self.users)

    @property
    def n_items(self):
        return len(self.items)

    @classmethod
    def create(cls, users: Vocabulary, items: Vocabulary, ratings: torch.Tensor) -> TrainingData:
        assert ratings.shape == (len(users), len(items))

        transposed = ratings.transpose(0, 1).to_sparse_csr()
        return cls(users, items, ratings, transposed)

    def to(self, device):
        """
        Move the training data to another device.
        """
        return self._replace(ui_rates=self.ui_rates.to(device), iu_rates=self.iu_rates.to(device))


class ALSBase(ABC, Component, Trainable):
    """
    Base class for ALS models.
    """

    features: int
    epochs: int
    reg: UITuple[float]
    rng: RNGInput
    save_user_features: bool

    users_: Vocabulary | None
    items_: Vocabulary
    user_features_: torch.Tensor | None
    item_features_: torch.Tensor

    @property
    @abstractmethod
    def logger(self) -> logging.Logger:  # pragma: no cover
        """
        Overridden in implementation to provide the logger.
        """
        ...

    def __init__(
        self,
        features: int,
        *,
        epochs: int = 10,
        reg: UITuple[float] | float | tuple[float, float] = 0.1,
        save_user_features: bool = True,
        rng: RNGInput = None,
    ):
        self.features = features
        self.epochs = epochs
        self.reg = UITuple.create(reg)
        self.rng = rng
        self.save_user_features = save_user_features

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "item_features_")

    @override
    def train(self, data: Dataset):
        """
        Run ALS to train a model.

        Args:
            ratings: the ratings data frame.
        """
        ensure_parallel_init()
        timer = util.Stopwatch()

        for algo in self.fit_iters(data, timer=timer):
            pass  # we just need to do the iterations

        if self.user_features_ is not None:
            self.logger.info(
                "trained model in %s (|P|=%f, |Q|=%f)",
                timer,
                torch.norm(self.user_features_, "fro"),
                torch.norm(self.item_features_, "fro"),
            )
        else:
            self.logger.info(
                "trained model in %s (|Q|=%f)",
                timer,
                torch.norm(self.item_features_, "fro"),
            )

    def fit_iters(self, data: Dataset, *, timer: util.Stopwatch | None = None) -> Iterator[Self]:
        """
        Run ALS to train a model, yielding after each iteration.

        Args:
            ratings: the ratings data frame.
        """
        if timer is None:
            timer = util.Stopwatch()

        train = self.prepare_data(data)
        self.users_ = train.users
        self.items_ = train.items

        self.initialize_params(train)

        if isinstance(self.reg, tuple):
            ureg, ireg = self.reg
        else:
            ureg = ireg = self.reg

        assert self.user_features_ is not None
        assert self.item_features_ is not None
        u_ctx = TrainContext.create(
            "user", train.ui_rates, self.user_features_, self.item_features_, ureg
        )
        i_ctx = TrainContext.create(
            "item", train.iu_rates, self.item_features_, self.user_features_, ireg
        )

        self.logger.info(
            "[%s] training biased MF model with ALS for %d features", timer, self.features
        )
        start = timer.elapsed()

        with item_progress("Training ALS", self.epochs) as epb:
            for epoch in range(self.epochs):
                epoch = epoch + 1

                du = self.als_half_epoch(epoch, u_ctx)
                self.logger.debug("[%s] finished user epoch %d", timer, epoch)

                di = self.als_half_epoch(epoch, i_ctx)
                self.logger.debug("[%s] finished item epoch %d", timer, epoch)

                self.logger.info(
                    "[%s] finished epoch %d (|ΔP|=%.3f, |ΔQ|=%.3f)", timer, epoch, du, di
                )
                epb.update()
                yield self

        if not self.save_user_features:
            self.user_features_ = None
            self.user_ = None

        end = timer.elapsed()
        self.logger.info(
            "[%s] trained %d epochs (%.1fs/epoch)",
            timer,
            self.epochs,
            (end - start) / self.epochs,
        )

    @abstractmethod
    def prepare_data(self, data: Dataset) -> TrainingData:  # pragma: no cover
        """
        Prepare data for training this model.  This takes in the ratings, and is
        supposed to do two things:

        -   Normalize or transform the rating/interaction data, as needed, for
            training.
        -   Store any parameters learned from the normalization (e.g. means) in
            the appropriate member variables.
        -   Return the training data object to use for model training.
        """
        ...

    def initialize_params(self, data: TrainingData):
        """
        Initialize the model parameters at the beginning of training.
        """
        rng = random_generator(self.rng)
        self.logger.debug("initializing item matrix")
        self.item_features_ = self.initial_params(data.n_items, self.features, rng)
        self.logger.debug("|Q|: %f", torch.norm(self.item_features_, "fro"))

        self.logger.debug("initializing user matrix")
        self.user_features_ = self.initial_params(data.n_users, self.features, rng)
        self.logger.debug("|P|: %f", torch.norm(self.user_features_, "fro"))

    @abstractmethod
    def initial_params(
        self, nrows: int, ncols: int, rng: np.random.Generator
    ) -> torch.Tensor:  # pragma: no cover
        """
        Compute initial parameter values of the specified shape.
        """
        ...

    @abstractmethod
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:  # pragma: no cover
        """
        Run one half of an ALS training epoch.
        """
        ...

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None and self.users_ is not None:
            user_num = self.users_.number(user_id, missing=None)

        u_offset = None
        u_feat = None
        if query.user_items is not None and len(query.user_items) > 0:
            u_feat, u_offset = self.new_user_embedding(user_num, query.user_items)

        if u_feat is None:
            if user_num is None or self.user_features_ is None:
                return ItemList(items, scores=np.nan)
            u_feat = self.user_features_[user_num, :]

        item_nums = items.numbers("torch", vocabulary=self.items_, missing="negative")
        item_mask = item_nums >= 0
        i_feats = self.item_features_[item_nums[item_mask], :]

        scores = torch.full((len(items),), np.nan, dtype=torch.float64)
        scores[item_mask] = i_feats @ u_feat

        results = ItemList(items, scores=scores)
        return self.finalize_scores(user_num, results, u_offset)

    @abstractmethod
    def new_user_embedding(
        self, user_num: int | None, items: ItemList
    ) -> tuple[torch.Tensor, float | None]:  # pragma: no cover
        """
        Generate an embedding for a user given their current ratings.
        """
        ...

    def finalize_scores(
        self, user_num: int | None, items: ItemList, user_bias: float | None
    ) -> ItemList:
        """
        Perform any final transformation of scores prior to returning them.
        """
        return items



"""
LensKit ranking (and list) metrics.
"""

from ._base import RankingMetricBase
from ._dcg import NDCG
from ._hit import Hit
from ._pr import Precision, Recall
from ._rbp import RBP
from ._recip import RecipRank

__all__ = [
    "RankingMetricBase",
    "Hit",
    "Precision",
    "Recall",
    "RecipRank",
    "NDCG",
    "RBP",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
from __future__ import annotations

import logging
from typing import Mapping

from lenskit.data import ID, GenericKey, ItemList, ItemListCollection
from lenskit.pipeline import Pipeline

from ._runner import BatchPipelineRunner

_logger = logging.getLogger(__name__)


def predict(
    pipeline: Pipeline,
    test: ItemListCollection[GenericKey] | Mapping[ID, ItemList],
    *,
    n_jobs: int | None = None,
    **kwargs,
) -> ItemListCollection[GenericKey]:
    """
    Convenience function to batch-generate rating predictions (or other per-item
    scores) from a pipeline.
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.predict()
    outs = runner.run(pipeline, test)
    return outs.output("predictions")  # type: ignore



import numpy as np
from typing_extensions import override

from lenskit.data.items import ItemList

from ._base import ListMetric, RankingMetricBase


class Precision(ListMetric, RankingMetricBase):
    """
    Compute recommendation precision.  This is computed as:

    .. math::
        \\frac{|L \\cap I_u^{\\mathrm{test}}|}{|L|}

    In the uncommon case that ``k`` is specified and ``len(recs) < k``, this metric uses
    ``len(recs)`` as the denominator.
    """

    @property
    def label(self):
        if self.k is not None:
            return f"Precision@{self.k}"
        else:
            return "Precision"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        nrecs = len(recs)
        if nrecs == 0:
            return np.nan

        items = recs.ids()
        ngood = np.isin(items, test.ids()).sum()
        return ngood / nrecs


class Recall(ListMetric, RankingMetricBase):
    """
    Compute recommendation recall.  This is computed as:

    .. math::
        \\frac{|L \\cap I_u^{\\mathrm{test}}|}{\\operatorname{min}\\{|I_u^{\\mathrm{test}}|, k\\}}
    """

    @property
    def label(self):
        if self.k is not None:
            return f"Recall@{self.k}"
        else:
            return "Recall"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)

        items = recs.ids()
        ngood = np.isin(items, test.ids()).sum()
        nrel = len(test)
        if self.k is not None and self.k < nrel:
            nrel = self.k
        return ngood / nrel



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from __future__ import annotations

import re
import warnings
from importlib import import_module
from types import GenericAlias, NoneType, UnionType
from typing import (  # type: ignore
    Generic,
    Protocol,
    TypeVar,
    Union,
    _GenericAlias,
    get_args,
    get_origin,
)

import numpy as np

T = TypeVar("T", covariant=True)


class TypecheckWarning(UserWarning):
    "Warnings about type-checking logic."

    pass


class Lazy(Protocol, Generic[T]):
    """
    Type for accepting lazy inputs from the pipeline runner.  If your function
    may or may not need one of its inputs, declare the type with this to only
    run it as needed:

    .. code:: python

        def my_component(input: str, backup: Lazy[str]) -> str:
            if input == 'invalid':
                return backup.get()
            else:
                return input
    """

    def get(self) -> T:
        """
        Get the value behind this lazy instance.
        """
        ...


def is_compatible_type(typ: type, *targets: type) -> bool:
    """
    Make a best-effort check whether a type is compatible with at least one
    target type. This function is limited by limitations of the Python type
    system and the effort required to (re-)write a full type checker.  It is
    written to be over-accepting instead of over-restrictive, so it can be used
    to reject clearly incompatible types without rejecting combinations it
    cannot properly check.

    Args:
        typ:
            The type to check.
        targets:
            One or more target types to check against.

    Returns:
        ``False`` if it is clear that the specified type is incompatible with
        all of the targets, and ``True`` otherwise.
    """
    for target in targets:
        # try a straight subclass check first, but gracefully handle incompatible types
        try:
            if issubclass(typ, target):
                return True
        except TypeError:
            pass

        if isinstance(target, (GenericAlias, _GenericAlias)):
            tcls = get_origin(target)
            # if we're matching a raw type against a generic, just check the origin
            if isinstance(typ, GenericAlias):
                warnings.warn(f"cannot type-check generic type {typ}", TypecheckWarning)
                cls = get_origin(typ)
                if issubclass(cls, tcls):  # type: ignore
                    return True
            elif isinstance(typ, type):
                print(typ, type(typ))
                if issubclass(typ, tcls):  # type: ignore
                    return True
        elif typ == int and isinstance(target, type) and issubclass(target, (float, complex)):  # noqa: E721
            return True
        elif typ == float and issubclass(target, complex):  # noqa: E721
            return True

    return False


def is_compatible_data(obj: object, *targets: type | TypeVar) -> bool:
    """
    Make a best-effort check whether a type is compatible with at least one
    target type. This function is limited by limitations of the Python type
    system and the effort required to (re-)write a full type checker.  It is
    written to be over-accepting instead of over-restrictive, so it can be used
    to reject clearly incompatible types without rejecting combinations it
    cannot properly check.

    Args:
        typ:
            The type to check.
        targets:
            One or more target types to check against.

    Returns:
        ``False`` if it is clear that the specified type is incompatible with
        all of the targets, and ``True`` otherwise.
    """
    for target in targets:
        # try a straight subclass check first, but gracefully handle incompatible types
        try:
            if isinstance(obj, target):
                return True
        except TypeError:
            pass

        origin = get_origin(target)
        if origin == UnionType or origin == Union:
            types = get_args(target)
            if is_compatible_data(obj, *types):
                return True
        elif isinstance(target, TypeVar):
            # is this quite correct?
            if target.__bound__ is None or isinstance(obj, target.__bound__):
                return True
        elif isinstance(target, (GenericAlias, _GenericAlias)):
            tcls = get_origin(target)
            if isinstance(obj, np.ndarray) and tcls == np.ndarray:
                # check for type compatibility
                _sz, dtw = get_args(target)
                (dt,) = get_args(dtw)
                if issubclass(obj.dtype.type, dt):
                    return True
            elif isinstance(tcls, type) and isinstance(obj, tcls):
                # this has holes, but is as close an approximation as we can get
                return True
        elif (
            isinstance(obj, int)
            and isinstance(target, type)
            and issubclass(target, (float, complex))
        ):  # noqa: E721
            return True
        elif isinstance(obj, float) and issubclass(target, complex):  # noqa: E721
            return True

    return False


def type_string(typ: type | None) -> str:
    """
    Compute a string representation of a type that is both resolvable and
    human-readable.  Type parameterizations are lost.
    """
    if typ is None or typ is NoneType:
        return "None"
    elif typ.__module__ == "builtins":
        return typ.__name__
    elif typ.__qualname__ == typ.__name__:
        return f"{typ.__module__}.{typ.__name__}"
    else:
        return f"{typ.__module__}:{typ.__qualname__}"


def parse_type_string(tstr: str) -> type:
    """
    Compute a string representation of a type that is both resolvable and
    human-readable.  Type parameterizations are lost.
    """
    if tstr == "None":
        return NoneType
    elif re.match(r"^\w+$", tstr):
        return __builtins__[tstr]
    else:
        if ":" in tstr:
            mod_name, typ_name = tstr.split(":", 1)
        else:
            # separate last element from module
            parts = re.match(r"(.*)\.(\w+)$", tstr)
            if not parts:
                raise ValueError(f"unparsable type string {tstr}")
            mod_name, typ_name = parts.groups()

        mod = import_module(mod_name)
        return getattr(mod, typ_name)



"""
Logging, progress, and resource records.
"""

import os
from typing import Any

import structlog

from .config import LoggingConfig
from .progress import Progress, item_progress, set_progress_impl
from .tasks import Task

__all__ = [
    "LoggingConfig",
    "Progress",
    "item_progress",
    "set_progress_impl",
    "Task",
    "get_logger",
    "trace",
]

get_logger = structlog.stdlib.get_logger
_trace_debug = os.environ.get("LK_TRACE", "no").lower() == "debug"


def trace(logger: structlog.stdlib.BoundLogger, *args: Any, **kwargs: Any):
    """
    Emit a trace-level message, if LensKit tracing is enabled.  Trace-level
    messages are more fine-grained than debug-level messages, and you usually
    don't want them.
    """
    meth = getattr(logger, "trace", None)
    if meth is not None:
        meth(*args, **kwargs)
    elif _trace_debug:
        logger.debug(*args, **kwargs)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data manipulation routines.
"""

# pyright: basic
from __future__ import annotations

import logging
from typing import Any, NamedTuple, Optional, TypeVar

import numpy as np
import pandas as pd
import scipy.sparse as sps
import torch
from numpy.typing import ArrayLike
from typing_extensions import override

from .dataset import Dataset, FieldError
from .items import ItemList
from .tables import NumpyUserItemTable, TorchUserItemTable
from .types import ID
from .vocab import Vocabulary

_log = logging.getLogger(__name__)

t = torch
M = TypeVar("M", "CSRStructure", sps.csr_array, sps.coo_array, sps.spmatrix, t.Tensor)


class CSRStructure(NamedTuple):
    """
    Representation of the compressed sparse row structure of a sparse matrix,
    without any data values.
    """

    rowptrs: np.ndarray
    colinds: np.ndarray
    shape: tuple[int, int]

    @property
    def nrows(self):
        return self.shape[0]

    @property
    def ncols(self):
        return self.shape[1]

    @property
    def nnz(self):
        return self.rowptrs[self.nrows]

    def extent(self, row: int) -> tuple[int, int]:
        return self.rowptrs[row], self.rowptrs[row + 1]

    def row_cs(self, row: int) -> np.ndarray:
        sp, ep = self.extent(row)
        return self.colinds[sp:ep]


class InteractionMatrix:
    """
    Internal helper class used by :class:`lenskit.data.Dataset` to store the
    user-item interaction matrix.  The data is stored simultaneously in CSR and
    COO format.  Most code has no need to interact with this class directly —
    :class:`~lenskit.data.Dataset` methods provide data in a range of formats.
    """

    n_obs: int
    n_users: int
    n_items: int

    user_nums: np.ndarray[int, np.dtype[np.int32]]
    "User (row) numbers."
    user_ptrs: np.ndarray[int, np.dtype[np.int32]]
    "User (row) offsets / pointers."
    item_nums: np.ndarray[int, np.dtype[np.int32]]
    "Item (column) numbers."
    ratings: Optional[np.ndarray[int, np.dtype[np.float32]]] = None
    "Rating values."
    timestamps: Optional[np.ndarray[int, np.dtype[np.int64]]] = None
    "Timestamps as 64-bit Unix timestamps."

    def __init__(
        self,
        users: ArrayLike,
        items: ArrayLike,
        ratings: Optional[ArrayLike],
        timestamps: Optional[ArrayLike],
        n_users: int,
        n_items: int,
    ):
        self.user_nums = np.asarray(users, np.int32)
        assert np.all(np.diff(self.user_nums) >= 0), "matrix data not sorted"
        self.item_nums = np.asarray(items, np.int32)
        if ratings is not None:
            self.ratings = np.asarray(ratings, np.float32)
        if timestamps is not None:
            self.timestamps = np.asarray(timestamps, np.int64)

        self.n_obs = len(self.user_nums)
        self.n_items = n_items
        self.n_users = n_users
        cp1 = np.zeros(self.n_users + 1, np.int32)
        np.add.at(cp1[1:], self.user_nums, 1)
        self.user_ptrs = cp1.cumsum(dtype=np.int32)
        if self.user_ptrs[-1] != len(self.user_nums):
            raise ValueError("mismatched counts and array sizes")

    @property
    def shape(self) -> tuple[int, int]:
        """
        The shape of the interaction matrix (rows x columns).
        """
        return (self.n_users, self.n_items)


class MatrixDataset(Dataset):
    """
    Dataset implementation using an in-memory rating or implicit-feedback matrix
    (with no duplicate interactions).

    .. note::
        Client code generally should not construct this class directly.  Instead
        use the various ``from_`` and ``load_`` functions in
        :mod:`lenskit.data`.
    """

    _users: Vocabulary
    "User ID vocabulary, to map between IDs and row numbers."
    _items: Vocabulary
    "Item ID vocabulary, to map between IDs and column or row numbers."
    _matrix: InteractionMatrix

    def __init__(self, users: Vocabulary, items: Vocabulary, interact_df: pd.DataFrame):
        """
        Construct a dataset.

        .. note::
            Client code generally should not call this constructor.  Instead use the
            various ``from_`` and ``load_`` functions in :mod:`lenskit.data`.
        """
        self._users = users
        self._items = items
        self._init_structures(interact_df)

    def _init_structures(self, df: pd.DataFrame):
        uno = self.users.numbers(df["user_id"])
        ino = self.items.numbers(df["item_id"])
        assert np.all(uno >= 0)
        assert np.all(ino >= 0)
        if np.any(df.duplicated(subset=["user_id", "item_id"])):
            raise RuntimeError("repeated ratings not yet supported")

        df = df.assign(user_num=uno, item_num=ino)

        _log.debug("sorting interaction table")
        df.sort_values(["user_num", "item_num"], ignore_index=True, inplace=True)
        _log.debug("rating data frame:\n%s", df)
        self._matrix = InteractionMatrix(
            uno,
            ino,
            df["rating"] if "rating" in df.columns else None,
            df["timestamp"] if "timestamp" in df.columns else None,
            self.user_count,
            self.item_count,
        )

    @property
    @override
    def items(self) -> Vocabulary:
        return self._items

    @property
    @override
    def users(self) -> Vocabulary:
        return self._users

    @override
    def count(self, what: str) -> int:
        match what:
            case "users":
                return self._users.size
            case "items":
                return self._items.size
            case "pairs" | "interactions" | "ratings":
                return self._matrix.n_obs
            case _:
                raise KeyError(f"unknown entity type {what}")

    @override
    def interaction_matrix(
        self,
        format: str,
        *,
        layout: str | None = None,
        legacy: bool = False,
        field: str | None = None,
        combine: str | None = None,
        original_ids: bool = False,
    ) -> Any:
        match format:
            case "structure":
                if layout and layout != "csr":
                    raise ValueError(f"unsupported layout {layout} for structure")
                if field:
                    raise ValueError("structure does not support fields")
                return self._int_mat_structure()
            case "pandas":
                if layout and layout != "coo":
                    raise ValueError(f"unsupported layout {layout} for Pandas")
                return self._int_mat_pandas(field, original_ids)
            case "scipy":
                return self._int_mat_scipy(field, layout, legacy)
            case "torch":
                return self._int_mat_torch(field, layout)
            case _:
                raise ValueError(f"unsupported format “{format}”")

    def _int_mat_structure(self) -> CSRStructure:
        return CSRStructure(self._matrix.user_ptrs, self._matrix.item_nums, self._matrix.shape)

    def _int_mat_pandas(self, field: str | None, original_ids: bool) -> pd.DataFrame:
        cols: dict[str, ArrayLike]
        if original_ids:
            cols = {
                "user_id": self.users.ids(self._matrix.user_nums),
                "item_id": self.items.ids(self._matrix.item_nums),
            }
        else:
            cols = {
                "user_num": self._matrix.user_nums,
                "item_num": self._matrix.item_nums,
            }
        if field == "all" or field == "rating":
            if self._matrix.ratings is not None:
                cols["rating"] = self._matrix.ratings
            else:
                cols["rating"] = np.ones(self._matrix.n_obs)
        elif field == "all" or field == "timestamp":
            if self._matrix.timestamps is None:
                raise FieldError("interaction", field)
            cols["timestamp"] = self._matrix.timestamps
        elif field and field != "all":
            raise FieldError("interaction", field)
        return pd.DataFrame(cols)

    def _int_mat_scipy(self, field: str | None, layout: str | None, legacy: bool):
        if field == "rating" and self._matrix.ratings is not None:
            data = self._matrix.ratings
        elif field is None or field == "rating":
            data = np.ones(self._matrix.n_obs, dtype="f4")
        elif field == "timestamp" and self._matrix.timestamps is not None:
            data = self._matrix.timestamps
        else:  # pragma nocover
            raise FieldError("interaction", field)

        shape = self._matrix.shape

        if layout is None:
            layout = "csr"
        match layout:
            case "csr":
                ctor = sps.csr_matrix if legacy else sps.csr_array
                return ctor((data, self._matrix.item_nums, self._matrix.user_ptrs), shape=shape)
            case "coo":
                ctor = sps.coo_matrix if legacy else sps.coo_array
                return ctor((data, (self._matrix.user_nums, self._matrix.item_nums)), shape=shape)
            case _:  # pragma nocover
                raise ValueError(f"unsupported layout {layout}")

    def _int_mat_torch(self, field: str | None, layout: str | None):
        if field == "rating" and self._matrix.ratings is not None:
            values = torch.from_numpy(self._matrix.ratings)
        elif field is None or field == "rating":
            values = torch.full([self._matrix.n_obs], 1.0, dtype=torch.float32)
        elif field == "timestamp" and self._matrix.timestamps is not None:
            values = torch.from_numpy(self._matrix.timestamps)
        else:  # pragma nocover
            raise FieldError("interaction", field)

        shape = self._matrix.shape

        if layout is None:
            layout = "csr"
        match layout:
            case "csr":
                return torch.sparse_csr_tensor(
                    torch.from_numpy(self._matrix.user_ptrs),
                    torch.from_numpy(self._matrix.item_nums),
                    values,
                    size=shape,
                )
            case "coo":
                indices = np.stack([self._matrix.user_nums, self._matrix.item_nums], dtype=np.int32)
                return torch.sparse_coo_tensor(
                    torch.from_numpy(indices),
                    values,
                    size=shape,
                ).coalesce()
            case _:  # pragma nocover
                raise ValueError(f"unsupported layout {layout}")

    @override
    def interaction_log(
        self,
        format: str,
        *,
        fields: str | list[str] | None = "all",
        original_ids: bool = False,
    ) -> Any:
        if fields == "all":
            fields = ["rating", "timestamp"]
        elif isinstance(fields, str):
            fields = [fields]
        elif fields is None:
            fields = []

        match format:
            case "pandas":
                return self._int_log_pandas(fields, original_ids)
            case "numpy":
                return self._int_log_numpy(fields)
            case "torch":
                return self._int_log_torch(fields)
            case _:
                raise ValueError(f"unsupported format “{format}”")

    def _int_log_pandas(self, fields: list[str], original_ids: bool):
        cols: dict[str, ArrayLike]
        if original_ids:
            cols = {
                "user_id": self.users.terms(self._matrix.user_nums),
                "item_id": self.items.terms(self._matrix.item_nums),
            }
        else:
            cols = {
                "user_num": self._matrix.user_nums,
                "item_num": self._matrix.item_nums,
            }
        if "rating" in fields and self._matrix.ratings is not None:
            cols["rating"] = self._matrix.ratings
        if "timestamp" in fields and self._matrix.timestamps is not None:
            cols["timestamp"] = self._matrix.timestamps
        return pd.DataFrame(cols)

    def _int_log_numpy(self, fields: list[str]) -> NumpyUserItemTable:
        tbl = NumpyUserItemTable(self._matrix.user_nums, self._matrix.item_nums)
        if "rating" in fields:
            tbl.ratings = self._matrix.ratings
        if "timestamp" in fields:
            tbl.timestamps = self._matrix.timestamps
        return tbl

    def _int_log_torch(self, fields: list[str]) -> TorchUserItemTable:
        tbl = TorchUserItemTable(
            torch.from_numpy(self._matrix.user_nums), torch.from_numpy(self._matrix.item_nums)
        )
        if "rating" in fields:
            tbl.ratings = torch.from_numpy(self._matrix.ratings)
        if "timestamp" in fields:
            tbl.timestamps = torch.from_numpy(self._matrix.timestamps)
        return tbl

    @override
    def user_row(
        self, user_id: ID | None = None, *, user_num: int | None = None
    ) -> ItemList | None:
        if user_num is None:
            if user_id is None:  # pragma: nocover
                raise ValueError("must provide one of user_id and user_num")

            user_num = self.users.number(user_id, "none")
            if user_num is None:
                return None

        elif user_id is not None:  # pragma: nocover
            raise ValueError("must provide one of user_id and user_num")

        sp = self._matrix.user_ptrs[user_num]
        ep = self._matrix.user_ptrs[user_num + 1]
        inums = self._matrix.item_nums[sp:ep]
        fields = {}
        if self._matrix.ratings is not None:
            fields["rating"] = self._matrix.ratings[sp:ep]
        if self._matrix.timestamps is not None:
            fields["timestamp"] = self._matrix.timestamps[sp:ep]
        return ItemList(item_nums=inums, vocabulary=self.items, **fields)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit pipeline abstraction.
"""

# pyright: strict
from __future__ import annotations

import typing
import warnings
from types import FunctionType, UnionType
from uuid import NAMESPACE_URL, uuid4, uuid5

import structlog
from typing_extensions import Any, Literal, Self, TypeAlias, TypeVar, cast, overload

from lenskit.data import Dataset

from . import config
from .components import (  # type: ignore # noqa: F401
    Component,
    Configurable,
    PipelineFunction,
    Trainable,
    fallback_on_none,
    instantiate_component,
)
from .config import PipelineConfig
from .nodes import ND, ComponentNode, InputNode, LiteralNode, Node
from .state import PipelineState
from .types import Lazy, parse_type_string

__all__ = [
    "Pipeline",
    "PipelineError",
    "PipelineWarning",
    "PipelineState",
    "Node",
    "PipelineFunction",
    "Configurable",
    "Trainable",
    "PipelineConfig",
    "Lazy",
    "Component",
    "RecPipelineBuilder",
    "topn_pipeline",
]

_log = structlog.stdlib.get_logger(__name__)

# common type var for quick use
T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
CloneMethod: TypeAlias = Literal["config", "pipeline-config"]

NAMESPACE_LITERAL_DATA = uuid5(NAMESPACE_URL, "https://ns.lenskit.org/literal-data/")


class PipelineError(Exception):
    """
    Pipeline configuration errors.

    .. note::

        This exception is only to note problems with the pipeline configuration
        and structure (e.g. circular dependencies).  Errors *running* the
        pipeline are raised as-is.
    """


class PipelineWarning(Warning):
    """
    Pipeline configuration and setup warnings.  We also emit warnings to the
    logger in many cases, but this allows critical ones to be visible even if
    the client code has not enabled logging.

    .. note::

        This warning is only to note problems with the pipeline configuration
        and structure (e.g. circular dependencies).  Errors *running* the
        pipeline are raised as-is.
    """


class Pipeline:
    """
    LensKit recommendation pipeline.  This is the core abstraction for using
    LensKit models and other components to produce recommendations in a useful
    way.  It allows you to wire together components in (mostly) abitrary graphs,
    train them on data, and serialize pipelines to disk for use elsewhere.

    If you have a scoring model and just want to generate recommenations with a
    default setup and minimal configuration, see :func:`topn_pipeline`.

    Args:
        name:
            A name for the pipeline.
        version:
            A numeric version for the pipeline.
    """

    name: str | None = None
    version: str | None = None

    _nodes: dict[str, Node[Any]]
    _aliases: dict[str, Node[Any]]
    _defaults: dict[str, Node[Any]]
    _components: dict[str, PipelineFunction[Any]]
    _hash: str | None = None
    _last: Node[Any] | None = None
    _anon_nodes: set[str]
    "Track generated node names."

    def __init__(self, name: str | None = None, version: str | None = None):
        self.name = name
        self.version = version
        self._nodes = {}
        self._aliases = {}
        self._defaults = {}
        self._components = {}
        self._anon_nodes = set()
        self._clear_caches()

    def meta(self, *, include_hash: bool = True) -> config.PipelineMeta:
        """
        Get the metadata (name, version, hash, etc.) for this pipeline without
        returning the whole config.

        Args:
            include_hash:
                Whether to include a configuration hash in the metadata.
        """
        meta = config.PipelineMeta(name=self.name, version=self.version)
        if include_hash:
            meta.hash = self.config_hash()
        return meta

    @property
    def nodes(self) -> list[Node[object]]:
        """
        Get the nodes in the pipeline graph.
        """
        return list(self._nodes.values())

    @overload
    def node(self, node: str, *, missing: Literal["error"] = "error") -> Node[object]: ...
    @overload
    def node(self, node: str, *, missing: Literal["none"] | None) -> Node[object] | None: ...
    @overload
    def node(self, node: Node[T]) -> Node[T]: ...
    def node(
        self, node: str | Node[Any], *, missing: Literal["error", "none"] | None = "error"
    ) -> Node[object] | None:
        """
        Get the pipeline node with the specified name.  If passed a node, it
        returns the node or fails if the node is not a member of the pipeline.

        Args:
            node:
                The name of the pipeline node to look up, or a node to check for
                membership.

        Returns:
            The pipeline node, if it exists.

        Raises:
            KeyError:
                The specified node does not exist.
        """
        if isinstance(node, Node):
            self._check_member_node(node)
            return node
        elif node in self._aliases:
            return self._aliases[node]
        elif node in self._nodes:
            return self._nodes[node]
        elif missing == "none" or missing is None:
            return None
        else:
            raise KeyError(f"node {node}")

    def create_input(self, name: str, *types: type[T] | None) -> Node[T]:
        """
        Create an input node for the pipeline.  Pipelines expect their inputs to
        be provided when they are run.

        Args:
            name:
                The name of the input.  The name must be unique in the pipeline
                (among both components and inputs).
            types:
                The allowable types of the input; input data can be of any
                specified type.  If ``None`` is among the allowed types, the
                input can be omitted.

        Returns:
            A pipeline node representing this input.

        Raises:
            ValueError:
                a node with the specified ``name`` already exists.
        """
        self._check_available_name(name)

        rts: set[type[T | None]] = set()
        for t in types:
            if t is None:
                rts.add(type(None))
            elif isinstance(t, UnionType):
                rts |= set(typing.get_args(t))
            else:
                rts.add(t)

        node = InputNode[Any](name, types=rts)
        self._nodes[name] = node
        self._clear_caches()
        return node

    def literal(self, value: T, *, name: str | None = None) -> LiteralNode[T]:
        """
        Create a literal node (a node with a fixed value).

        .. note::
            Literal nodes cannot be serialized witih :meth:`get_config` or
            :meth:`save_config`.
        """
        if name is None:
            name = str(uuid4())
            self._anon_nodes.add(name)
        node = LiteralNode(name, value, types=set([type(value)]))
        self._nodes[name] = node
        self._clear_caches()
        return node

    def set_default(self, name: str, node: Node[Any] | object) -> None:
        """
        Set the default wiring for a component input.  Components that declare
        an input parameter with the specified ``name`` but no configured input
        will be wired to this node.

        This is intended to be used for things like wiring up `user` parameters
        to semi-automatically receive the target user's identity and history.

        Args:
            name:
                The name of the parameter to set a default for.
            node:
                The node or literal value to wire to this parameter.
        """
        if not isinstance(node, Node):
            node = self.literal(node)
        self._defaults[name] = node
        self._clear_caches()

    def get_default(self, name: str) -> Node[Any] | None:
        """
        Get the default wiring for an input name.
        """
        return self._defaults.get(name, None)

    def alias(self, alias: str, node: Node[Any] | str) -> None:
        """
        Create an alias for a node.  After aliasing, the node can be retrieved
        from :meth:`node` using either its original name or its alias.

        Args:
            alias:
                The alias to add to the node.
            node:
                The node (or node name) to alias.

        Raises:
            ValueError:
                if the alias is already used as an alias or node name.
        """
        node = self.node(node)
        self._check_available_name(alias)
        self._aliases[alias] = node
        self._clear_caches()

    def add_component(
        self, name: str, obj: Component[ND] | PipelineFunction[ND], **inputs: Node[Any] | object
    ) -> Node[ND]:
        """
        Add a component and connect it into the graph.

        Args:
            name:
                The name of the component in the pipeline.  The name must be
                unique in the pipeline (among both components and inputs).
            obj:
                The component itself.
            inputs:
                The component's input wiring.  See :ref:`pipeline-connections`
                for details.

        Returns:
            The node representing this component in the pipeline.
        """
        self._check_available_name(name)

        node = ComponentNode(name, obj)
        self._nodes[name] = node
        self._components[name] = obj

        self.connect(node, **inputs)

        self._clear_caches()
        self._last = node
        return node

    def replace_component(
        self,
        name: str | Node[ND],
        obj: Component[ND] | PipelineFunction[ND],
        **inputs: Node[Any] | object,
    ) -> Node[ND]:
        """
        Replace a component in the graph.  The new component must have a type
        that is compatible with the old component.  The old component's input
        connections will be replaced (as the new component may have different
        inputs), but any connections that use the old component to supply an
        input will use the new component instead.
        """
        if isinstance(name, Node):
            name = name.name

        node = ComponentNode(name, obj)
        self._nodes[name] = node
        self._components[name] = obj

        self.connect(node, **inputs)

        self._clear_caches()
        return node

    def connect(self, obj: str | Node[Any], **inputs: Node[Any] | str | object):
        """
        Provide additional input connections for a component that has already
        been added.  See :ref:`pipeline-connections` for details.

        Args:
            obj:
                The name or node of the component to wire.
            inputs:
                The component's input wiring.  For each keyword argument in the
                component's function signature, that argument can be provided
                here with an input that the pipeline will provide to that
                argument of the component when the pipeline is run.
        """
        if isinstance(obj, Node):
            node = obj
        else:
            node = self.node(obj)
        if not isinstance(node, ComponentNode):
            raise TypeError(f"only component nodes can be wired, not {node}")

        for k, n in inputs.items():
            if isinstance(n, Node):
                n = cast(Node[Any], n)
                self._check_member_node(n)
                node.connections[k] = n.name
            else:
                lit = self.literal(n)
                node.connections[k] = lit.name

        self._clear_caches()

    def component_configs(self) -> dict[str, dict[str, Any]]:
        """
        Get the configurations for the components.  This is the configurations
        only, it does not include pipeline inputs or wiring.
        """
        return {
            name: comp.get_config()
            for (name, comp) in self._components.items()
            if isinstance(comp, Configurable)
        }

    def clone(self, how: CloneMethod = "config") -> Pipeline:
        """
        Clone the pipeline, optionally including trained parameters.

        The ``how`` parameter controls how the pipeline is cloned, and what is
        available in the clone pipeline.  It can be one of the following values:

        ``"config"``
            Create fresh component instances using the configurations of the
            components in this pipeline.  When applied to a trained pipeline,
            the clone does **not** have the original's learned parameters. This
            is the default clone method.
        ``"pipeline-config"``
            Round-trip the entire pipeline through :meth:`get_config` and
            :meth:`from_config`.

        Args:
            how:
                The mechanism to use for cloning the pipeline.

        Returns:
            A new pipeline with the same components and wiring, but fresh
            instances created by round-tripping the configuration.
        """
        if how == "pipeline-config":
            cfg = self.get_config()
            return self.from_config(cfg)
        elif how != "config":  # pragma: nocover
            raise NotImplementedError("only 'config' cloning is currently supported")

        clone = Pipeline()

        for node in self.nodes:
            match node:
                case InputNode(name, types=types):
                    if types is None:
                        types = set[type]()
                    clone.create_input(name, *types)
                case LiteralNode(name, value):
                    clone._nodes[name] = LiteralNode(name, value)
                case ComponentNode(name, comp, _inputs, wiring):
                    if isinstance(comp, FunctionType):
                        comp = comp
                    elif isinstance(comp, Configurable):
                        comp = comp.__class__.from_config(comp.get_config())  # type: ignore
                    else:
                        comp = comp.__class__()  # type: ignore
                    cn = clone.add_component(node.name, comp)  # type: ignore
                    for wn, wt in wiring.items():
                        clone.connect(cn, **{wn: clone.node(wt)})
                case _:  # pragma: nocover
                    raise RuntimeError(f"invalid node {node}")

        for n, t in self._aliases.items():
            clone.alias(n, t.name)

        for n, t in self._defaults.items():
            clone.set_default(n, clone.node(t.name))

        return clone

    def get_config(self, *, include_hash: bool = True) -> PipelineConfig:
        """
        Get this pipeline's configuration for serialization.  The configuration
        consists of all inputs and components along with their configurations
        and input connections.  It can be serialized to disk (in JSON, YAML, or
        a similar format) to save a pipeline.

        The configuration does **not** include any trained parameter values,
        although the configuration may include things such as paths to
        checkpoints to load such parameters, depending on the design of the
        components in the pipeline.

        .. note::
            Literal nodes (from :meth:`literal`, or literal values wired to
            inputs) cannot be serialized, and this method will fail if they
            are present in the pipeline.
        """
        meta = self.meta(include_hash=False)
        cfg = PipelineConfig(meta=meta)

        # We map anonymous nodes to hash-based names for stability.  If we ever
        # allow anonymous components, this will need to be adjusted to maintain
        # component ordering, but it works for now since only literals can be
        # anonymous. First handle the anonymous nodes, so we have that mapping:
        remapped: dict[str, str] = {}
        for an in self._anon_nodes:
            node = self._nodes.get(an, None)
            match node:
                case None:
                    # skip nodes that no longer exist
                    continue
                case LiteralNode(name, value):
                    lit = config.PipelineLiteral.represent(value)
                    sname = str(uuid5(NAMESPACE_LITERAL_DATA, lit.model_dump_json()))
                    _log.debug("renamed anonymous node %s to %s", name, sname)
                    remapped[name] = sname
                    cfg.literals[sname] = lit
                case _:
                    # the pipeline only generates anonymous literal nodes right now
                    raise RuntimeError(f"unexpected anonymous node {node}")

        # Now we go over all named nodes and add them to the config:
        for node in self.nodes:
            if node.name in remapped:
                continue

            match node:
                case InputNode():
                    cfg.inputs.append(config.PipelineInput.from_node(node))
                case LiteralNode(name, value):
                    cfg.literals[name] = config.PipelineLiteral.represent(value)
                case ComponentNode(name):
                    cfg.components[name] = config.PipelineComponent.from_node(node, remapped)
                case _:  # pragma: nocover
                    raise RuntimeError(f"invalid node {node}")

        cfg.aliases = {a: t.name for (a, t) in self._aliases.items()}
        cfg.defaults = {n: t.name for (n, t) in self._defaults.items()}

        if include_hash:
            cfg.meta.hash = config.hash_config(cfg)

        return cfg

    def config_hash(self) -> str:
        """
        Get a hash of the pipeline's configuration to uniquely identify it for
        logging, version control, or other purposes.

        The hash format and algorithm are not guaranteed, but is stable within a
        LensKit version.  For the same version of LensKit and component code,
        the same configuration will produce the same hash, so long as there are
        no literal nodes.  Literal nodes will *usually* hash consistently, but
        since literals other than basic JSON values are hashed by pickling, hash
        stability depends on the stability of the pickle bytestream.

        In LensKit 2025.1, the configuration hash is computed by computing the
        JSON serialization of the pipeline configuration *without* a hash and
        returning the hex-encoded SHA256 hash of that configuration.
        """
        if self._hash is None:
            # get the config *without* a hash
            cfg = self.get_config(include_hash=False)
            self._hash = config.hash_config(cfg)
        return self._hash

    @classmethod
    def from_config(cls, config: object) -> Self:
        cfg = PipelineConfig.model_validate(config)
        pipe = cls()
        for inpt in cfg.inputs:
            types: list[type[Any] | None] = []
            if inpt.types is not None:
                types += [parse_type_string(t) for t in inpt.types]
            pipe.create_input(inpt.name, *types)

        # we now add the components and other nodes in multiple passes to ensure
        # that nodes are available before they are wired (since `connect` can
        # introduce out-of-order dependencies).

        # pass 1: add literals
        for name, data in cfg.literals.items():
            pipe.literal(data.decode(), name=name)

        # pass 2: add components
        to_wire: list[config.PipelineComponent] = []
        for name, comp in cfg.components.items():
            if comp.code.startswith("@"):
                # ignore special nodes in first pass
                continue

            obj = instantiate_component(comp.code, comp.config)
            pipe.add_component(name, obj)
            to_wire.append(comp)

        # pass 3: wiring
        for name, comp in cfg.components.items():
            if isinstance(comp.inputs, dict):
                inputs = {n: pipe.node(t) for (n, t) in comp.inputs.items()}
                pipe.connect(name, **inputs)
            elif not comp.code.startswith("@"):
                raise PipelineError(f"component {name} inputs must be dict, not list")

        # pass 4: aliases
        for n, t in cfg.aliases.items():
            pipe.alias(n, t)

        # pass 5: defaults
        for n, t in cfg.defaults.items():
            pipe.set_default(n, pipe.node(t))

        if cfg.meta.hash is not None:
            h2 = pipe.config_hash()
            if h2 != cfg.meta.hash:
                _log.warning("loaded pipeline does not match hash")
                warnings.warn("loaded pipeline config does not match hash", PipelineWarning)

        return pipe

    def train(self, data: Dataset, *, retrain: bool = True) -> None:
        """
        Trains the pipeline's trainable components (those implementing the
        :class:`TrainableComponent` interface) on some training data.

        Args:
            data:
                The dataset to train on.
            retrain:
                Whether to re-train components that have already been trained.
        """
        for comp in self._components.values():
            if hasattr(comp, "train"):
                comp = cast(Trainable, comp)
                if comp.is_trained and not retrain:
                    _log.debug("component %s already trained", comp)
                    continue
                _log.info("training %s", comp)
                comp.train(data)
            else:
                _log.debug("component %s does not need training", comp)

    @overload
    def run(self, /, **kwargs: object) -> object: ...
    @overload
    def run(self, node: str, /, **kwargs: object) -> object: ...
    @overload
    def run(self, n1: str, n2: str, /, *nrest: str, **kwargs: object) -> tuple[object]: ...
    @overload
    def run(self, node: Node[T], /, **kwargs: object) -> T: ...
    @overload
    def run(self, n1: Node[T1], n2: Node[T2], /, **kwargs: object) -> tuple[T1, T2]: ...
    @overload
    def run(
        self, n1: Node[T1], n2: Node[T2], n3: Node[T3], /, **kwargs: object
    ) -> tuple[T1, T2, T3]: ...
    @overload
    def run(
        self, n1: Node[T1], n2: Node[T2], n3: Node[T3], n4: Node[T4], /, **kwargs: object
    ) -> tuple[T1, T2, T3, T4]: ...
    @overload
    def run(
        self,
        n1: Node[T1],
        n2: Node[T2],
        n3: Node[T3],
        n4: Node[T4],
        n5: Node[T5],
        /,
        **kwargs: object,
    ) -> tuple[T1, T2, T3, T4, T5]: ...
    def run(self, *nodes: str | Node[Any], **kwargs: object) -> object:
        """
        Run the pipeline and obtain the return value(s) of one or more of its
        components.  See :ref:`pipeline-execution` for details of the pipeline
        execution model.

        Args:
            nodes:
                The component(s) to run.
            kwargs:
                The pipeline's inputs, as defined with :meth:`create_input`.

        Returns:
            The pipeline result.  If zero or one nodes are specified, the result
            is returned as-is. If multiple nodes are specified, their results
            are returned in a tuple.

        Raises:
            PipelineError:
                when there is a pipeline configuration error (e.g. a cycle).
            ValueError:
                when one or more required inputs are missing.
            TypeError:
                when one or more required inputs has an incompatible type.
            other:
                exceptions thrown by components are passed through.
        """
        if not nodes:
            if self._last is None:  # pragma: nocover
                raise PipelineError("pipeline has no components")
            nodes = (self._last,)
        state = self.run_all(*nodes, **kwargs)
        results = [state[self.node(n).name] for n in nodes]

        if len(results) > 1:
            return tuple(results)
        else:
            return results[0]

    def run_all(self, *nodes: str | Node[Any], **kwargs: object) -> PipelineState:
        """
        Run all nodes in the pipeline, or all nodes required to fulfill the
        requested node, and return a mapping with the full pipeline state (the
        data attached to each node). This is useful in cases where client code
        needs to be able to inspect the data at arbitrary steps of the pipeline.
        It differs from :meth:`run` in two ways:

        1.  It returns the data from all nodes as a mapping (dictionary-like
            object), not just the specified nodes as a tuple.
        2.  If no nodes are specified, it runs *all* nodes instead of only the
            last node.  This has the consequence of running nodes that are not
            required to fulfill the last node (such scenarios typically result
            from using :meth:`use_first_of`).

        Args:
            nodes:
                The nodes to run, as positional arguments (if no nodes are
                specified, this method runs all nodes).
            kwargs:
                The inputs.

        Returns:
            The full pipeline state, with :attr:`~PipelineState.default` set to
            the last node specified (either the last node in `nodes`, or the
            last node added to the pipeline).
        """
        from .runner import PipelineRunner

        runner = PipelineRunner(self, kwargs)
        node_list = [self.node(n) for n in nodes]
        _log.debug("running pipeline", name=self.name, nodes=[n.name for n in node_list])
        if not node_list:
            node_list = self.nodes

        last = None
        for node in node_list:
            runner.run(node)
            last = node.name

        return PipelineState(
            runner.state,
            {a: t.name for (a, t) in self._aliases.items()},
            default=last,
            meta=self.meta(),
        )

    def use_first_of(self, name: str, input: Node[T | None], fallback: Node[T]) -> Node[T]:
        """
        Ergonomic method to create a new node that returns the result of its
        ``input`` if it is provided and not ``None``, and otherwise returns the
        result of ``fallback``.  This method is used for things like filling in
        optional pipeline inputs.  For example, if you want the pipeline to take
        candidate items through an ``items`` input, but look them up from the
        user's history and the training data if ``items`` is not supplied, you
        would do:

        .. code:: python

            pipe = Pipeline()
            # allow candidate items to be optionally specified
            items = pipe.create_input('items', list[EntityId], None)
            # find candidates from the training data (optional)
            lookup_candidates = pipe.add_component(
                'select-candidates', UnratedTrainingItemsCandidateSelector(),
                user=history,
            )
            # if the client provided items as a pipeline input, use those; otherwise
            # use the candidate selector we just configured.
            candidates = pipe.use_first_of('candidates', items, lookup_candidates)

        .. note::

            This method does not distinguish between an input being unspecified
            and explicitly specified as ``None``.

        .. note::

            This method does *not* implement item-level fallbacks, only
            fallbacks at the level of entire results.  For item-level score
            fallbacks, see :class:`~lenskit.basic.FallbackScorer`.

        .. note::
            If one of the fallback elements is a component ``A`` that depends on
            another component or input ``B``, and ``B`` is missing or returns
            ``None`` such that ``A`` would usually fail, then ``A`` will be
            skipped and the fallback will move on to the next node. This works
            with arbitrarily-deep transitive chains.

        Args:
            name:
                The name of the node.
            input:
                The node to use, if it is available.
            fallback:
                The node to use if ``input`` does not provide a value.
        """
        return self.add_component(name, fallback_on_none, input=input, fallback=fallback)

    def _check_available_name(self, name: str) -> None:
        if name in self._nodes or name in self._aliases:
            raise ValueError(f"pipeline already has node {name}")

    def _check_member_node(self, node: Node[Any]) -> None:
        nw = self._nodes.get(node.name)
        if nw is not node:
            raise PipelineError(f"node {node} not in pipeline")

    def _clear_caches(self):
        if "_hash" in self.__dict__:
            del self._hash


# remaining re-exports
from .common import RecPipelineBuilder, topn_pipeline  # type: ignore # noqa: E402



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Warning and error classes and general LensKit diagnostic code.
"""


class DataWarning(UserWarning):
    """
    Warning raised for detectable problems with input data.
    """

    pass


class ConfigWarning(UserWarning):
    """
    Warning raised for detectable problems with algorithm configurations.
    """

    pass



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
from __future__ import annotations

import logging
from typing import Sequence

from lenskit.data import ID, ItemListCollection, UserIDKey
from lenskit.pipeline import Pipeline

from ._runner import BatchPipelineRunner

_logger = logging.getLogger(__name__)


def recommend(
    pipeline: Pipeline,
    users: Sequence[ID | UserIDKey],
    n: int | None = None,
    candidates=None,
    *,
    n_jobs: int | None = None,
    **kwargs,
) -> ItemListCollection[UserIDKey]:
    """
    Convenience function to batch-generate recommendations from a pipeline.
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.recommend(n=n)
    outs = runner.run(pipeline, users)
    return outs.output("recommendations")
